package no.kolkinn.subsym.project2;

import no.kolkinn.cbr_dovre.cbr.Caze;
import no.kolkinn.cbr_dovre.cbr.CazeFileReader;
import no.kolkinn.cbr_dovre.models.Attributes;
import no.kolkinn.cbr_dovre.models.Model1;
import no.kolkinn.cbr_dovre.models.Model2;
import no.kolkinn.cbr_dovre.models.SimilarityFunctions;
import no.kolkinn.cbr_dovre.tests.Tests;
import no.kolkinn.subsym.project2.problems.CbrProblem;
import no.kolkinn.subsym.project2.util.StringTools;
import no.kolkinn.subsym.project2.util.TriFunction;

import java.io.File;
import java.io.IOException;
import java.nio.file.Paths;
import java.text.MessageFormat;
import java.text.ParseException;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.time.temporal.TemporalUnit;
import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.function.BiFunction;
import java.util.logging.FileHandler;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Created by Mikael on 25.09.2016.
 */
public class ConfigurationTester {
    private final String logDir;
    private final Logger logger;

    public ConfigurationTester(String logDirectory) throws IOException {
        if(logDirectory != null && !logDirectory.isEmpty()) {

            if(logDirectory.endsWith("/")) {
                this.logDir = logDirectory.substring(0, logDirectory.length() - 1);
            } else {
                this.logDir = logDirectory;
            }

            // Create logdir if it doesn't exist
            new File(this.logDir).mkdirs();
        } else {
            this.logDir = Paths.get(".").toAbsolutePath().normalize().toString();
        }

        LocalDateTime dateTime = LocalDateTime.now();
        String logDate = dateTime.format(DateTimeFormatter.ofPattern("dd-MM-uuuu_HH-mm", Locale.getDefault()));
        logger = Logger.getAnonymousLogger(); // Make each logger unique to the instance
        logger.addHandler(new FileHandler(this.logDir + "/configuration_test_" + logDate + ".%u.log"));
        logger.setLevel(Level.CONFIG);
    }

    public void testWithTestSet(Test test) {
        test.logger = logger;

        if(!(test.problemSettings instanceof CbrSettings)) {
            logger.severe(
                    "test.problemSettings was not an instance of CbrSettings, cannot perform test.\n"
                    + "test.problemSettings: " + test.problemSettings.getClass());
            return;
        }

        CbrSettings cbrSettings = (CbrSettings) test.problemSettings;
        EvolutionaryAlgorithm.Results results = test.getResults();

        Map<String, Double> weights;
        if(cbrSettings.weights != null) {
            weights = cbrSettings.weights;
        } else if(results != null && results.bestIndividual != null) {
            weights = ((CbrProblem.CbrPhenotype) results.bestIndividual).getWeights();
        } else {
            weights = null;
        }

        SimilarityFunctions similarityFunctions;
        if(cbrSettings.useModel2) {
            similarityFunctions = new Model2(weights, cbrSettings.useSimpleKNN);
        } else {
            similarityFunctions = new Model1(weights, cbrSettings.useSimpleKNN);
        }
        try {
            Tests.runTests(logger, cbrSettings.getTrainingCases(), cbrSettings.getTestCases(), cbrSettings.kNN,
                           1, cbrSettings.onlyOtherLocations, similarityFunctions);
        } catch(IOException | ParseException e) {
            logger.log(Level.SEVERE, "Exception when getting cases.", e);
            logger.log(Level.SEVERE, "Cause of the exception:", e.getCause());
        } catch(InterruptedException | ExecutionException e) {
            logger.log(Level.SEVERE, "Exception when running test(s).", e);
            logger.log(Level.SEVERE, "Cause of the exception:", e.getCause());
        }
        logger.warning("The test above was run with the following settings:\n"
                       + CbrProblem.explainWhetherSimpleKNN(cbrSettings.useSimpleKNN)
                       + "\nWeights:\n"
                       + weights + '\n'
                       + test.toString());
    }

    public Test runTests(List<Test> tests) {
        for(Test test : tests) {
            try {
                test.logger = logger;
                test.runTest();
                logger.info(test.toString());
            } catch(IOException | ParseException e) {
                logger.log(Level.SEVERE, "Error when running test: " + test.toString(), e);
                logger.log(Level.SEVERE, "Cause of the exception:", e.getCause());
            }
        }

        Optional<Test> best = tests.stream().max(Comparator.comparingDouble(Test::getFitness));
        if(best.isPresent()) {
            logger.warning("Best configuration:\n"
                           + best.get().toString());
            return best.get();
        } else {
            logger.severe("Could not find best configuration!");
            return null;
        }
    }

    public static class Test {
        EASettings eaSettings;
        ProblemSettings problemSettings;
        AdultSelectionSettings adultSelectionSettings;
        ParentSelectionSettings parentSelectionSettings;
        private Logger logger = Logger.getAnonymousLogger();
        private EvolutionaryAlgorithm.Results results;

        public Test(EASettings eaSettings, ProblemSettings problemSettings,
                    AdultSelectionSettings adultSelectionSettings,
                    ParentSelectionSettings parentSelectionSettings) {
            this.eaSettings = eaSettings;
            this.problemSettings = problemSettings;
            this.adultSelectionSettings = adultSelectionSettings;
            this.parentSelectionSettings = parentSelectionSettings;
        }

        public EvolutionaryAlgorithm.Results getResults() {
            return results;
        }

        @Override
        public String toString() {
            return "Test [fitness: " + getFitness() + "] {" +
                   StringTools.indent(
                           "\n" + eaSettings +
                           ",\n" + problemSettings +
                           ",\n" + adultSelectionSettings +
                           ",\n" + parentSelectionSettings +
                           ",\nGenotype:\n" + StringTools.indent(
                                   results == null ? null : results.bestIndividual == null ? null : results.bestIndividual.toString())
                                     ) +
                   "\n}";
        }

        public double getFitness() {
            if(results == null || results.bestIndividual == null) {
                logger.warning(
                        "The call to Test.getFitness() could not return a proper value as it was not yet set. Returning Double.NaN");
                return Double.NaN;
            }
            return results.bestIndividual.getFitness();
        }

        public void runTest() throws IOException, ParseException {
            EvolutionaryAlgorithm ea = new EvolutionaryAlgorithm(problemSettings.getProblem(),
                                                                 adultSelectionSettings.getAdultSelector(),
                                                                 parentSelectionSettings.getParentSelector());
            ea.setLoggerMethod((generationNumber, bestYoungFitness, averageYoungFitness, youngStandardDeviation,
                                bestYoungIndividual) -> logger.config(MessageFormat.format(
                    "Generation: {0}, best young fitness: {1}, average young fitness {2}, young standard deviation: {3}"
                    + "\nBest young individual: {4}",
                    generationNumber, bestYoungFitness, averageYoungFitness, youngStandardDeviation,
                    bestYoungIndividual)));

            results = ea.evolutionaryLoop(null,
                                          eaSettings.populationSize,
                                          eaSettings.crossoverRate,
                                          eaSettings.overProductionSize,
                                          eaSettings.generationStop,
                                          eaSettings.fitnessStop,
                                          eaSettings.roundsWithoutImprovementStop,
                                          eaSettings.elitismAsChildren);
        }
    }

    public static class EASettings {
        public int populationSize;
        public int elitismAsChildren;
        public double crossoverRate;
        public int overProductionSize;
        public int generationStop;
        public double fitnessStop;
        public int roundsWithoutImprovementStop;

        public EASettings(int populationSize, int elitismAsChildren, double crossoverRate, int overProductionSize,
                          int generationStop, double fitnessStop, int roundsWithoutImprovementStop) {
            this.populationSize = populationSize;
            this.elitismAsChildren = elitismAsChildren;
            this.crossoverRate = crossoverRate;
            this.overProductionSize = overProductionSize;
            this.generationStop = generationStop;
            this.fitnessStop = fitnessStop;
            this.roundsWithoutImprovementStop = roundsWithoutImprovementStop;
        }

        @Override
        public String toString() {
            return "EASettings{" +
                   "\n\tpopulationSize=" + populationSize +
                   ",\n\telitismAsChildren=" + elitismAsChildren +
                   ",\n\tcrossoverRate=" + crossoverRate +
                   ",\n\toverProductionSize=" + overProductionSize +
                   ",\n\tgenerationStop=" + generationStop +
                   ",\n\tfitnessStop=" + fitnessStop +
                   ",\n\troundsWithoutImprovementStop=" + roundsWithoutImprovementStop +
                   "\n}";
        }
    }

    public static abstract class ParentSelectionSettings {
        public abstract BiFunction<List<Problem.Phenotype>, Integer, List<Problem.Phenotype>> getParentSelector();

        @Override
        public String toString() {
            return "ParentSelectionSettings{}";
        }
    }

    public static class SigmaScalingSettings extends ParentSelectionSettings {

        @Override
        public BiFunction<List<Problem.Phenotype>, Integer, List<Problem.Phenotype>> getParentSelector() {
            return ParentSelection::SigmaScaling;
        }

        @Override
        public String toString() {
            return "SigmaScalingSettings{}";
        }
    }

    public static class FitnessProportionateSettings extends ParentSelectionSettings {

        @Override
        public BiFunction<List<Problem.Phenotype>, Integer, List<Problem.Phenotype>> getParentSelector() {
            return ParentSelection::FitnessProportionate;
        }

        @Override
        public String toString() {
            return "FitnessProportionateSettings{}";
        }
    }

    public static class TournamentSelectionSettings extends ParentSelectionSettings {
        public int numCompetitors;
        public double randomWinnerProbability;

        public TournamentSelectionSettings(int numCompetitors, double randomWinnerProbability) {
            this.numCompetitors = numCompetitors;
            this.randomWinnerProbability = randomWinnerProbability;
        }

        @Override
        public BiFunction<List<Problem.Phenotype>, Integer, List<Problem.Phenotype>> getParentSelector() {
            return (adults, numParents) -> ParentSelection.Tournament(adults, numParents, numCompetitors,
                                                                      randomWinnerProbability);
        }

        @Override
        public String toString() {
            return "TournamentSelectionSettings{" +
                   "\n\tnumCompetitors=" + numCompetitors +
                   ",\n\trandomWinnerProbability=" + randomWinnerProbability +
                   "\n}";
        }
    }

    public static class RankSelectionSettings extends ParentSelectionSettings {
        public double min;

        public RankSelectionSettings(double min) {
            this.min = min;
        }

        @Override
        public BiFunction<List<Problem.Phenotype>, Integer, List<Problem.Phenotype>> getParentSelector() {
            return (adults, numParents) -> ParentSelection.Rank(adults, numParents, min);
        }

        @Override
        public String toString() {
            return "RankSelectionSettings{" +
                   "\n\tmin=" + min +
                   "\n}";
        }
    }

    public static class AdultSelectionSettings {
        public boolean mixGenerations;
        public int elitismAsParents;

        public AdultSelectionSettings(boolean mixGenerations, int elitismAsParents) {
            this.mixGenerations = mixGenerations;
            this.elitismAsParents = elitismAsParents;
        }

        public TriFunction<List<Problem.Phenotype>, List<Problem.Phenotype>, Integer, List<Problem.Phenotype>> getAdultSelector() {
            if(mixGenerations) {
                return (oldAdults, youngAdults, numToKeep) -> AdultSelection.greedy(
                        AdultSelection.mixGenerations(oldAdults, youngAdults), numToKeep);
            } else {
                return (oldAdults, youngAdults, numToKeep) -> AdultSelection.greedy(
                        AdultSelection.elitist(oldAdults, youngAdults, elitismAsParents),
                        numToKeep);
            }
        }

        @Override
        public String toString() {
            return "AdultSelectionSettings{" +
                   "\n\tmixGenerations=" + mixGenerations +
                   ",\n\telitismAsParents=" + elitismAsParents +
                   "\n}";
        }
    }

    public static abstract class ProblemSettings {
        public abstract Problem getProblem() throws IOException, ParseException;

        @Override
        public String toString() {
            return "ProblemSettings{}";
        }
    }

    public static class CbrSettings extends ProblemSettings {
        public final int kNN;
        public int crossoverPoints;
        public boolean uniformCrossover;
        public double mutationProbability;
        public boolean mutatePerComponent;
        public int bitsPerWeights;
        public boolean useModel2;
        public boolean useSimpleKNN;
        public int numTestsPerGeneration;
        public boolean prune = false;
        public long groupDistance = 24;
        public TemporalUnit timeUnit = ChronoUnit.HOURS;
        public Map<String, Double> weights;
        public final TrainingTestSplit trainingTestSplit;
        public boolean onlyOtherLocations;

        public CbrSettings(int crossoverPoints, boolean uniformCrossover, double mutationProbability,
                           boolean mutatePerComponent, int bitsPerWeights, boolean useModel2, boolean useSimpleKNN,
                           int kNN, int numTestsPerGeneration, Map<String, Double> weights, boolean prune,
                           long groupDistance, TemporalUnit timeUnit, TrainingTestSplit trainingTestSplit,
                           boolean onlyOtherLocations) {
            this.crossoverPoints = crossoverPoints;
            this.uniformCrossover = uniformCrossover;
            this.mutationProbability = mutationProbability;
            this.mutatePerComponent = mutatePerComponent;
            this.bitsPerWeights = bitsPerWeights;
            this.useModel2 = useModel2;
            this.useSimpleKNN = useSimpleKNN;
            this.kNN = kNN;
            this.numTestsPerGeneration = numTestsPerGeneration;
            this.weights = weights;
            this.prune = prune;
            this.groupDistance = groupDistance;
            this.timeUnit = timeUnit;
            this.trainingTestSplit = trainingTestSplit;
            this.onlyOtherLocations = onlyOtherLocations;
        }

        @Override
        public Problem getProblem() throws IOException, ParseException {
            Set<Caze> cazes = getTrainingCases();
            CbrProblem problem = new CbrProblem(crossoverPoints, uniformCrossover, mutatePerComponent,
                                                mutationProbability,
                                                bitsPerWeights, useModel2, cazes, numTestsPerGeneration, useSimpleKNN,
                                                kNN);

            if(prune) {
                problem.setPruneTrainingCases(true, groupDistance, timeUnit);
            }

            return problem;
        }

        @Override
        public String toString() {
            return "CbrSettings{" +
                   "\n\tkNN=" + kNN +
                   ",\n\tcrossoverPoints=" + crossoverPoints +
                   ",\n\tuniformCrossover=" + uniformCrossover +
                   ",\n\tmutationProbability=" + mutationProbability +
                   ",\n\tmutatePerComponent=" + mutatePerComponent +
                   ",\n\tbitsPerWeights=" + bitsPerWeights +
                   ",\n\tuseModel2=" + useModel2 +
                   ",\n\tuseSimpleKNN=" + useSimpleKNN +
                   ",\n\tnumTestsPerGeneration=" + numTestsPerGeneration +
                   ",\n\tonlyOtherLocations=" + onlyOtherLocations +
                   ",\n\tprune=" + prune +
                   ",\n\tgroupDistance=" + groupDistance +
                   ",\n\ttimeUnit=" + timeUnit +
                   ",\n\tweights=" + (weights == null ? "null" : weights.toString()) +
                   ",\n" + StringTools.indent(trainingTestSplit.toString()) +
                   "\n}";
        }

        public Set<Caze> getTrainingCases() throws IOException, ParseException {
            return trainingTestSplit.getTrainingCases();
        }

        public Set<Caze> getTestCases() throws IOException, ParseException {
            return trainingTestSplit.getTestCases();
        }

        public void setTrainingCases(Set<Caze> cases) {
            trainingTestSplit.setTrainingCases(cases);
        }

        public void setTestCases(Set<Caze> cases) {
            trainingTestSplit.setTestCases(cases);
        }

        public static class TrainingTestSplit {
            private File trainingCasesFile;
            private File testCasesFile;
            private Set<Caze> trainingCases = null;
            private Set<Caze> testCases = null;

            public TrainingTestSplit(String trainingCasesFilePath, String testCasesFilePath) {
                this.trainingCasesFile = new File(trainingCasesFilePath);
                this.testCasesFile = new File(testCasesFilePath);
            }

            public TrainingTestSplit(File trainingCasesFile, File testCasesFile) {
                this.trainingCasesFile = trainingCasesFile;
                this.testCasesFile = testCasesFile;
            }

            /**
             * Searches the given directory for files named trainingset[i].csv and testset[i].csv,
             * and returns a list of TrainingTestSplits.
             * [i] is replaced with a number, e.g. trainingset0.csv and testset0.csv, trainingset1.csv and testset1.csv.
             */
            public static List<TrainingTestSplit> getTrainingTestSplits(String directoryPath) {
                List<TrainingTestSplit> trainingTestSplits = new ArrayList<>();

                File directory = new File(directoryPath);
                File[] listOfFiles = directory.listFiles();

                Pattern filePattern = Pattern.compile("(training|test)set(\\d+)\\.csv");

                Map<String, List<File>> splitFiles = new HashMap<>();

                if(listOfFiles != null) {
                    for(File file : listOfFiles) {
                        Matcher matcher = filePattern.matcher(file.getName());
                        if(matcher.matches()) {
                            String key = matcher.group(2);
                            if(splitFiles.get(key) == null) {
                                List<File> list = new ArrayList<>(2);
                                list.add(file);
                                splitFiles.put(key, list);
                            } else {
                                splitFiles.get(key).add(file);
                            }
                        }
                    }

                    for(List<File> split : splitFiles.values()) {
                        // If both trainingset[key].csv and testset[key].csv exist
                        if(split.size() == 2) {
                            File trainingSet = null, testSet = null;
                            for(File file : split) {
                                if(file.getName().startsWith("training")) {
                                    trainingSet = file;
                                } else {
                                    testSet = file;
                                }
                            }
                            trainingTestSplits.add(new TrainingTestSplit(trainingSet, testSet));
                        }
                    }
                }

                return trainingTestSplits;
            }

            public Set<Caze> getTestCases() throws IOException, ParseException {
                if(testCases == null) {
                    testCases = CazeFileReader.readFile(testCasesFile, Attributes.ALL_LOCATIONS);
                }
                return testCases;
            }

            public Set<Caze> getTrainingCases() throws IOException, ParseException {
                if(trainingCases == null) {
                    trainingCases = CazeFileReader.readFile(trainingCasesFile, Attributes.ALL_LOCATIONS);
                }
                return trainingCases;
            }

            @Override
            public String toString() {
                return "TrainingTestSplit{" +
                       "\n\ttrainingCasesFilePath='" + trainingCasesFile.getAbsolutePath() + '\'' +
                       ",\n\ttestCasesFilePath='" + testCasesFile.getAbsolutePath() + '\'' +
                       "\n}";
            }

            public void setTrainingCases(Set<Caze> trainingCases) {
                this.trainingCases = trainingCases;
            }

            public void setTestCases(Set<Caze> testCases) {
                this.testCases = testCases;
            }
        }
    }
}
