package no.kolkinn.subsym.project2;

import no.kolkinn.subsym.project2.util.SeptiConsumer;
import no.kolkinn.subsym.project2.util.TriFunction;
import org.apache.commons.math3.stat.descriptive.SummaryStatistics;

import java.rmi.UnexpectedException;
import java.util.*;
import java.util.concurrent.ThreadLocalRandom;
import java.util.function.BiFunction;
import java.util.stream.Collectors;

/**
 * Created by Mikael on 17.02.2016.
 */
public class EvolutionaryAlgorithm {

    private final Problem problem;
    private final TriFunction<List<Problem.Phenotype>, List<Problem.Phenotype>, Integer, List<Problem.Phenotype>> selectAdults;
    private final BiFunction<List<Problem.Phenotype>, Integer, List<Problem.Phenotype>> selectParents;
    private SeptiConsumer<Integer, Double, Double, Double, Double> plotData = null;
    private SeptiConsumer<Integer, Double, Double, Double, String> logData = null;
    private volatile boolean stop;
    private Results results;

    /**
     * @param problem        problem definition
     * @param adultSelector  method for selecting which adults should proceed to parent selection. Pattern: {@code adultSelector(oldAdults, youngAdults, numAdultsToSelect) returns selectedAdults}
     * @param parentSelector method for selecting parents from adult pool. Pattern: {@code parentSelector(adults, numParentsToSelect) returns selectedParents}
     */
    public EvolutionaryAlgorithm(Problem problem,
                                 TriFunction<List<Problem.Phenotype>, List<Problem.Phenotype>, Integer, List<Problem.Phenotype>> adultSelector,
                                 BiFunction<List<Problem.Phenotype>, Integer, List<Problem.Phenotype>> parentSelector) {

        this.problem = problem;
        this.selectAdults = adultSelector;
        this.selectParents = parentSelector;
        this.results = null;
    }

    public Results evolutionaryLoop(List<Problem.Genotype> initialPopulation, int populationSize,
                                    double crossoverRate, int overProductionSize, int generationStop,
                                    double fitnessStop, int roundsWithoutImprovementStop, int elitism) throws UnexpectedException {

        stop = false;

        if(crossoverRate < 1e-10) {
            crossoverRate = 0.;
        }

        if(overProductionSize < populationSize) {
            overProductionSize = populationSize;
        }

        if(generationStop <= 0) {
            generationStop = Integer.MAX_VALUE;
        }

        if(fitnessStop < 1e-10) {
            fitnessStop = Double.POSITIVE_INFINITY;
        }

        if(roundsWithoutImprovementStop <= 0) {
            roundsWithoutImprovementStop = Integer.MAX_VALUE;
        }

        Problem.Phenotype bestIndividual = null;
        List<Problem.Genotype> children = new ArrayList<>(overProductionSize);
        List<Problem.Phenotype> youngAdults = new ArrayList<>(overProductionSize);
        List<Problem.Phenotype> adults = new ArrayList<>(populationSize);
        List<Problem.Phenotype> parents;
        results = new Results();

        int generationNumber = 0;
        int roundsWithoutImprovement = 0;

        // Generate initial population if empty
        if(initialPopulation == null || initialPopulation.isEmpty()) {
            for(int i = 0; i < populationSize; i++) {
                children.add(problem.generateGenotype());
            }
        } else {
            children.addAll(initialPopulation);
        }

        if(children.isEmpty()) {
            throw new UnexpectedException("Could not generate initial genotypes.");
        }

        while(true) {
            // Make preparations for the next generation if necessary
            problem.prepareNewGeneration(generationNumber);

            // Develop children into young adults
            youngAdults.clear();
            youngAdults.addAll(children.stream().map(Problem.Genotype::developPhenotype).collect(Collectors.toList()));


            if(youngAdults.isEmpty()) {
                throw new UnexpectedException("Could not develop genotypes into phenotypes.");
            }

            // Test fitness of young adults
            Problem.Phenotype bestYoungIndividual = Collections.max(youngAdults, Comparator.comparingDouble(
                    Problem.Phenotype::getFitness));
            if(bestIndividual == null || bestYoungIndividual.getFitness() > bestIndividual.getFitness()) {
                roundsWithoutImprovement = 0;
                bestIndividual = bestYoungIndividual;
            } else {
                roundsWithoutImprovement++;
            }

            // Update results
            results.bestIndividual = bestIndividual;
            results.children = children;
            results.youngAdults = youngAdults;
            results.adults = adults;

            // Plotting
            if(plotData != null || logData != null) {
                SummaryStatistics statistics = new SummaryStatistics();
                youngAdults.forEach(phenotype -> statistics.addValue(phenotype.getFitness()));

                double absoluteBestFitness = bestIndividual.getFitness();
                double bestYoungFitness = bestYoungIndividual.getFitness();
                double averageYoungFitness = youngAdults.stream().mapToDouble(
                        Problem.Phenotype::getFitness).average().getAsDouble();
                double youngStandardDeviation = statistics.getStandardDeviation();


                if(plotData != null) {
                    plotData.apply(generationNumber,
                                   absoluteBestFitness,
                                   bestYoungFitness,
                                   averageYoungFitness,
                                   youngStandardDeviation);
                }
                if(logData != null) {
                    logData.apply(generationNumber, bestYoungFitness, averageYoungFitness, youngStandardDeviation,
                                  bestYoungIndividual.toString());
                }
            }

            // Stopping criteria
            if(stop || generationNumber >= generationStop || bestIndividual.getFitness() + 1e-10 >= fitnessStop || roundsWithoutImprovement >= roundsWithoutImprovementStop) {
                break;
            }

            // Adult selection
            adults = selectAdults.apply(adults, youngAdults, populationSize);

            if(adults.isEmpty()) {
                throw new UnexpectedException("Adult selection turned out empty.");
            }

            if(stop) {
                break;
            }

            // Parent selection
            parents = selectParents.apply(adults, overProductionSize);

            if(parents.isEmpty()) {
                throw new UnexpectedException("Parent selection turned out empty.");
            }

            if(stop) {
                break;
            }

            // Reproduction
            children = reproduce(parents, crossoverRate);

            if(children.isEmpty()) {
                throw new UnexpectedException("Could not produce children.");
            }

            // Elitism
            children.addAll(AdultSelection.elitist(adults, Collections.emptyList(), elitism).stream().map(
                    Problem.Phenotype::getGenotype).collect(Collectors.toCollection(ArrayList<Problem.Genotype>::new)));

            generationNumber++;
        }

        problem.stop();

        return results;
    }

    public void stopEvolutionaryLoop() {
        this.stop = true;
    }

    @Override
    protected void finalize() throws Throwable {
        super.finalize();

        problem.stop();
    }

    public void setPlotterMethod(SeptiConsumer<Integer, Double, Double, Double, Double> plotterMethod) {
        plotData = plotterMethod;
    }

    public void setLoggerMethod(SeptiConsumer<Integer, Double, Double, Double, String> loggerMethod) {
        logData = loggerMethod;
    }

    private List<Problem.Genotype> reproduce(List<Problem.Phenotype> parents, double crossoverRate) {
        List<Problem.Genotype> children = new ArrayList<>(parents.size());

        Collections.shuffle(parents, ThreadLocalRandom.current());
        Iterator<Problem.Phenotype> iterator = parents.iterator();
        Problem.Genotype genotypeA, genotypeB;

        // Crossover
        // TODO: 19.02.2016 Have an option to try not to cross identical parents
        while(iterator.hasNext()) {
            genotypeA = iterator.next().getGenotype();

            if(iterator.hasNext() && ThreadLocalRandom.current().nextDouble() < crossoverRate) {
                genotypeB = iterator.next().getGenotype();
                Collections.addAll(children, genotypeA.crossover(genotypeB));
            } else {
                children.add(genotypeA.copy());
            }
        }

        // Mutation
        children.forEach(Problem.Genotype::mutate);

        return children;
    }

    public static class Results {
        public Problem.Phenotype bestIndividual = null;
        public List<Problem.Phenotype> youngAdults = null;
        public List<Problem.Phenotype> adults = null;
        public List<Problem.Genotype> children = null;
    }
}
