package no.kolkinn.subsym.project2;

import org.apache.commons.math3.stat.descriptive.SummaryStatistics;

import java.util.*;
import java.util.concurrent.ThreadLocalRandom;
import java.util.stream.Collectors;

/**
 * Created by Mikael on 17.02.2016.
 */
public class ParentSelection {
    /**
     * Uses numParents evenly spaced out pointers and spins the wheel.
     * @return List with the selected parents
     */
    private static List<Problem.Phenotype> rouletteWheel(List<PhenotypeWrapper> adults, int numParents) {
        double pointerInterval = ThreadLocalRandom.current().nextDouble() / numParents;
        boolean[] usedPointerIndexes = new boolean[numParents];
        List<PhenotypeWrapper> chosenParents = new ArrayList<>(numParents);
        double minSelect = 0.0;
        Collections.shuffle(adults, ThreadLocalRandom.current());
        for(PhenotypeWrapper adult : adults) {
            double maxSelect = minSelect + adult.scaledFitness;

            assert maxSelect <= 1.0 + 1E-1 : "maxSelect was significantly higher than 1.0";

            for(int i = 0; i < numParents; i++) {
                // To prevent pointer pointing at multiple adults,
                // only add once per pointer
                if(!usedPointerIndexes[i]) {
                    double pointer = (double) i / numParents + pointerInterval;

                    if(pointer >= minSelect && pointer < maxSelect) {
                        chosenParents.add(adult);
                        usedPointerIndexes[i] = true;
                    }
                }
            }

            minSelect = maxSelect;
        }

        assert chosenParents.size() == numParents : "chosenParents.size() = " + chosenParents.size() + ", should be " + numParents;

        return chosenParents.stream().map(PhenotypeWrapper::getPhenotype).collect(Collectors.toList());
    }

    private static void NormalizeScaledFitness(List<PhenotypeWrapper> adults){
        double scaledFitnessSum = adults.stream().mapToDouble(adult -> adult.scaledFitness).sum();

        if(Double.isNaN(scaledFitnessSum)) {
            throw new RuntimeException("Sum of scaled fitness values was NaN.");
        }

        for(PhenotypeWrapper adult : adults) {
            adult.scaledFitness /= scaledFitnessSum;
        }

        double test = adults.stream().mapToDouble(adult -> adult.scaledFitness).sum();
        assert Math.abs(test - 1.0) < 1E-10 : String.format("Scaled fitness values did not sum to 1.0. Was: %f",
                                                                        test);
    }

    public static List<Problem.Phenotype> FitnessProportionate(List<Problem.Phenotype> adults, int numParents) {
        if(adults.size() < 2) {
            return OneOrZeroAdults(adults, numParents);
        }

        List<PhenotypeWrapper> wrappedAdults = adults.stream().map(PhenotypeWrapper::new).collect(Collectors.toList());

        double averageFitness = wrappedAdults.stream().mapToDouble(
                PhenotypeWrapper::getFitness).average().getAsDouble();

        // Scale
        for(PhenotypeWrapper adult : wrappedAdults) {
            // Prevent divide by zero
            if(Math.abs(averageFitness) < 1E-13) {
                adult.scaledFitness = 1.0;
            } else {
                adult.scaledFitness = adult.getFitness()/averageFitness;
            }

            if(Double.isNaN(adult.scaledFitness)) {
                throw new RuntimeException("Scaled fitness was NaN.");
            }
        }

        // Normalize
        NormalizeScaledFitness(wrappedAdults);

        return rouletteWheel(wrappedAdults, numParents);
    }

    public static List<Problem.Phenotype> SigmaScaling(List<Problem.Phenotype> adults, int numParents) {
        // Not enough samples to calculate standard deviation
        if(adults.size() < 2) {
            return OneOrZeroAdults(adults, numParents);
        }

        List<PhenotypeWrapper> wrappedAdults = adults.stream().map(PhenotypeWrapper::new).collect(Collectors.toList());

        double averageFitness = wrappedAdults.stream().mapToDouble(
                PhenotypeWrapper::getFitness).average().getAsDouble();
        SummaryStatistics statistics = new SummaryStatistics();
        for(PhenotypeWrapper adult : wrappedAdults) {
            statistics.addValue(adult.getFitness());
        }
        double sdFitness = statistics.getStandardDeviation();

        // Scale
        for(PhenotypeWrapper adult : wrappedAdults) {
            // Prevent divide by zero
            if(Math.abs(sdFitness) < 1E-13) {
                adult.scaledFitness = 1.0;
            }
            else {
                adult.scaledFitness = 1.0 + (adult.getFitness() - averageFitness)/(2*sdFitness);
            }

            if(Double.isNaN(adult.scaledFitness)) {
                throw new RuntimeException("Scaled fitness was NaN.");
            }
        }

        // Normalize
        NormalizeScaledFitness(wrappedAdults);
        assert Math.abs(wrappedAdults.stream().mapToDouble(adult -> adult.scaledFitness).sum() - 1.0) < 1E-10;

        return rouletteWheel(wrappedAdults, numParents);
    }

    private static List<Problem.Phenotype> OneOrZeroAdults(List<Problem.Phenotype> adults, int numParents) {
        if(adults.size() == 0) {
            return new ArrayList<>();
        }

        List<Problem.Phenotype> parents = new ArrayList<>(numParents);
        Problem.Phenotype adult = adults.get(0);
        for(int i = 0; i < numParents; i++) {
            parents.add(adult.getGenotype().copy().developPhenotype());
        }

        return parents;
    }

    public static List<Problem.Phenotype> Tournament(List<Problem.Phenotype> adults, int numParents, int numCompetitors, double randomWinnerProbability) {
        if(randomWinnerProbability > 1 || randomWinnerProbability < 0) {
            throw new IllegalArgumentException("randomWinnerProbability was " + randomWinnerProbability + ". Must be 0 <= x <= 1");
        }

        if(numCompetitors > adults.size() || numCompetitors < 1) {
            throw new IllegalArgumentException("numCompetitors was " + numCompetitors + ". The K-parameter (number of competitors in a local tournament) was higher than the total population or less than 1. Initialize TournamentConfigurator with different parameters!");
        }

        List<Problem.Phenotype> winners = new ArrayList<>();

        for(int i = 0; i < numParents; i++) {
            Collections.shuffle(adults, ThreadLocalRandom.current());
            List<Problem.Phenotype> competitors = new ArrayList<>(numCompetitors);
            competitors.addAll(adults.subList(0, numCompetitors));

            if(ThreadLocalRandom.current().nextDouble() < randomWinnerProbability) {
                winners.add(competitors.get(ThreadLocalRandom.current().nextInt(competitors.size())));
            }
            else {
                winners.add(Collections.max(competitors, Comparator.comparingDouble(Problem.Phenotype::getFitness)));
            }
        }

        return winners;
    }

    public static List<Problem.Phenotype> Rank(List<Problem.Phenotype> adults, int numParents, double min) {
        if (adults.size() < 2) {
            return OneOrZeroAdults(adults, numParents);
        }

        List<PhenotypeWrapper> wrappedAdults = adults.stream().map(PhenotypeWrapper::new).collect(Collectors.toList());

        if(min > 1 || min < 0) {
            throw new IllegalArgumentException("min was " + min + ". Must be 0 <= min <= 1");
        }

        double max = 2. - min;


        int populationSize = wrappedAdults.size();

        Collections.sort(wrappedAdults, Comparator.comparingDouble(PhenotypeWrapper::getFitness));
        for(int i = 0; i < populationSize; i++) {
            wrappedAdults.get(i).scaledFitness = min + (max - min) * i / (populationSize - 1);
        }

        // Normalize
        NormalizeScaledFitness(wrappedAdults);

        return rouletteWheel(wrappedAdults, numParents);
    }

    static class PhenotypeWrapper {

        private final Problem.Phenotype phenotype;
        public double scaledFitness;

        public PhenotypeWrapper(Problem.Phenotype phenotype) {
            this.phenotype = phenotype;
        }

        public double getFitness() {
            return phenotype.getFitness();
        }

        public Problem.Phenotype getPhenotype() {
            return phenotype;
        }
    }

}
