package no.kolkinn.subsym.project2.controllers;

import javafx.fxml.FXML;
import javafx.scene.control.CheckBox;
import javafx.scene.control.TextField;
import no.kolkinn.subsym.project2.EvolutionaryAlgorithm;
import no.kolkinn.subsym.project2.Problem;
import no.kolkinn.subsym.project2.util.StringTools;

import java.rmi.UnexpectedException;
import java.util.ArrayList;
import java.util.List;

public class MainController {
    @FXML TextField averageingRounds;
    @FXML
    CheckBox generatePlots;
    @FXML
    CheckBox generateLogs;
    @FXML
    EASettingsController eaSettingsController;
    @FXML
    ParentSelectionController parentSelectionController;
    @FXML
    AdultSelectionController adultSelectionController;
    @FXML
    ProblemSettingsController problemSettingsController;
    @FXML
    ChartController chartController;
    @FXML
    ResultsController resultsController;

    EvolutionaryAlgorithm ea;
    List<EvolutionaryAlgorithm> running = new ArrayList<>();
    EvolutionaryAlgorithm.Results eaResults;

    public void start() {
        runEvolutionaryLoop(null, getNumAverageingRounds());
    }

    public void continueRun() {
        runEvolutionaryLoop(eaResults.children, getNumAverageingRounds());
    }

    public void stop() {
        running.forEach(EvolutionaryAlgorithm::stopEvolutionaryLoop);
        running.clear();
    }

    private void runEvolutionaryLoop(List<Problem.Genotype> initialPopulation, int averageingRounds) {
        if(averageingRounds < 1) {
            averageingRounds = 1;
        }
        final int finalAverageingRounds = averageingRounds;


        Thread eaThread = new Thread(() -> {
            Problem problem;
            try {
                problem = problemSettingsController.getProblemDefinition();
                ea = new EvolutionaryAlgorithm(problem,
                                               adultSelectionController.getSelectionMethod(),
                                               parentSelectionController.getSelectionMethod());
            } catch(UnexpectedException e) {
                e.printStackTrace();
                return;
            }

            if(generatePlots.isSelected()) {
                chartController.reset();
                if(finalAverageingRounds > 1) {
                    ea.setPlotterMethod(new AveragePlotting(eaSettingsController.getGenerationStop() + 1)::plotAverage);
                } else {
                    ea.setPlotterMethod(chartController::add);
                }
            }
            if(generateLogs.isSelected()) {
                System.out.println(
                        "\n\n\n-----------------------------------------------------------------------------------------------------------\n");
                ea.setLoggerMethod(this::logData);
            }

            running.add(ea);
            for(int i = 0; i < finalAverageingRounds; i++) {
                try {
                    eaResults = eaSettingsController.evolutionaryLoop(ea, initialPopulation);
                } catch(UnexpectedException e) {
                    e.printStackTrace();
                }
                resultsController.printResults(eaResults);
            }
            running.remove(ea);

            problemSettingsController.visualizeResults(problem, eaResults);

        }, "EA-thread");
        eaThread.setUncaughtExceptionHandler((th, ex) -> {
            System.err.println("Uncaught exception in EA-thread: " + ex);
            ex.printStackTrace();
        });
        eaThread.setDaemon(true);
        eaThread.start();
    }

    private int getNumAverageingRounds() {
        String text = averageingRounds.getText();
        if(text.trim().isEmpty()) {
            return 1;
        } else {
            return Integer.valueOf(text);
        }
    }

    private void logData(int generation, double bestFitness, double averageFitness, double standardDeviation, String phenotype) {
        System.out.format("Generation: % 5d \t Best: % 3.3f \t Average: % 3.3f \t SD: % 3.3f\n"
                          + "%s\n",
                          generation, bestFitness, averageFitness, standardDeviation, StringTools.indent(phenotype));
    }
    private class AveragePlotting {
        private final int[] runsPerGeneration;
        private final double[] averageAbsoluteBestFitness;
        private final double[] averageBestYoungFitness;
        private final double[] averageAverageYoungFitness;
        private final double[] averageYoungStandardDeviation;

        public AveragePlotting(int maxGenerations) {
            runsPerGeneration = new int[maxGenerations];
            averageAbsoluteBestFitness = new double[maxGenerations];
            averageBestYoungFitness = new double[maxGenerations];
            averageAverageYoungFitness = new double[maxGenerations];
            averageYoungStandardDeviation = new double[maxGenerations];
        }

        void plotAverage(int generation, double absoluteBestFitness, double bestYoungFitness, double averageYoungFitness, double youngStandardDeviation) {
            runsPerGeneration[generation]++;
            averageAbsoluteBestFitness[generation] += absoluteBestFitness;
            averageBestYoungFitness[generation] += bestYoungFitness;
            averageAverageYoungFitness[generation] += averageYoungFitness;
            averageYoungStandardDeviation[generation] += youngStandardDeviation;

            if(generation == 0) {
                chartController.reset();
            }
            chartController.add(generation,
                                averageAbsoluteBestFitness[generation] / runsPerGeneration[generation],
                                averageBestYoungFitness[generation] / runsPerGeneration[generation],
                                averageAverageYoungFitness[generation] / runsPerGeneration[generation],
                                averageYoungStandardDeviation[generation] / runsPerGeneration[generation]);
        }
    }
}
