package no.kolkinn.subsym.project2.controllers;

import javafx.beans.binding.Bindings;
import javafx.beans.binding.BooleanBinding;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.fxml.Initializable;
import javafx.scene.control.ChoiceBox;
import javafx.scene.control.TextField;
import javafx.scene.layout.VBox;
import no.kolkinn.subsym.project2.util.InputValidations;
import no.kolkinn.subsym.project2.ParentSelection;
import no.kolkinn.subsym.project2.Problem;

import java.net.URL;
import java.util.List;
import java.util.ResourceBundle;
import java.util.function.BiFunction;

/**
 * Created by Mikael on 20.02.2016.
 */
public class ParentSelectionController implements Initializable {

    public ChoiceBox<String> parentSelectorChoiceBox;

    private static final String FITNESS_PROPORTIONATE = "Fitness proportionate",
            SIGMA_SCALING = "Sigma scaling",
            TOURNAMENT = "Tournament selection",
            RANK = "Rank selection";

    @FXML
    private VBox tournamentConfigurationVBox;
    @FXML
    private VBox rankConfigurationVBox;
    @FXML
    TextField tournamentNumCompetitors;
    @FXML
    private TextField tournamentRandomWinnerProb;
    @FXML
    private TextField rankMin;

    private final ObservableList<String> selectionMethods = FXCollections.observableArrayList(
            SIGMA_SCALING,
            FITNESS_PROPORTIONATE,
            TOURNAMENT,
            RANK);

    @Override
    public void initialize(URL location, ResourceBundle resources) {
        BooleanBinding tournamentVisibility = Bindings.equal(TOURNAMENT, parentSelectorChoiceBox.valueProperty());
        BooleanBinding rankVisibility = Bindings.equal(RANK, parentSelectorChoiceBox.valueProperty());

        tournamentConfigurationVBox.managedProperty().bind(tournamentConfigurationVBox.visibleProperty());
        tournamentConfigurationVBox.visibleProperty().bind(tournamentVisibility);
        rankConfigurationVBox.managedProperty().bind(rankConfigurationVBox.visibleProperty());
        rankConfigurationVBox.visibleProperty().bind(rankVisibility);

        parentSelectorChoiceBox.setItems(selectionMethods);
        parentSelectorChoiceBox.setValue(selectionMethods.get(0));
    }

    public BiFunction<List<Problem.Phenotype>, Integer, List<Problem.Phenotype>> getSelectionMethod() {
        return getSelectionMethod(parentSelectorChoiceBox.getValue());
    }

    private BiFunction<List<Problem.Phenotype>, Integer, List<Problem.Phenotype>> getSelectionMethod(String name) {
        BiFunction<List<Problem.Phenotype>, Integer, List<Problem.Phenotype>> selectionMethod = null;
        boolean error = false;

        switch(name) {
            case FITNESS_PROPORTIONATE:
                selectionMethod = ParentSelection::FitnessProportionate;
                break;

            case SIGMA_SCALING:
                selectionMethod = ParentSelection::SigmaScaling;
                break;

            case TOURNAMENT:
                error |= !InputValidations.isInteger(tournamentNumCompetitors, 1, Integer.MAX_VALUE);
                error |= !InputValidations.isDouble(tournamentRandomWinnerProb, 0., 1.);

                if(!error) {
                    int numCompetitors = Integer.valueOf(tournamentNumCompetitors.getText());
                    double randomWinnerProb = Double.valueOf(tournamentRandomWinnerProb.getText());
                    selectionMethod = (phenotypes, integer) ->
                            ParentSelection.Tournament(phenotypes, integer, numCompetitors, randomWinnerProb);
                }

                break;

            case RANK:
                error = !InputValidations.isDouble(rankMin, 0., 1.);

                if(!error) {
                    double min = Double.valueOf(rankMin.getText());
                    selectionMethod = (phenotypes, integer) ->
                            ParentSelection.Rank(phenotypes, integer, min);
                }

                break;

            default:
                selectionMethod = (phenotypes, integer) -> phenotypes;
        }

        if(error) {
            throw new IllegalArgumentException("Illegal argument in ParentSelectionController");
        } else {
            return selectionMethod;
        }
    }
}
