package no.kolkinn.subsym.project2.controllers.problems;

import javafx.application.Platform;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.Initializable;
import javafx.scene.control.Button;
import javafx.scene.control.CheckBox;
import javafx.scene.control.TextField;
import javafx.stage.FileChooser;
import no.kolkinn.cbr_dovre.cbr.Caze;
import no.kolkinn.cbr_dovre.cbr.CazeFileReader;
import no.kolkinn.cbr_dovre.models.Attributes;
import no.kolkinn.cbr_dovre.models.Model1;
import no.kolkinn.cbr_dovre.models.Model2;
import no.kolkinn.cbr_dovre.tests.Tests;
import no.kolkinn.subsym.project2.EvolutionaryAlgorithm;
import no.kolkinn.subsym.project2.problems.CbrProblem;
import no.kolkinn.subsym.project2.util.InputValidations;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.rmi.UnexpectedException;
import java.text.ParseException;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.concurrent.ExecutionException;
import java.util.logging.Logger;

/**
 * Created by Mikael on 30.03.2016.
 */
public class CbrController implements Initializable {

    private final static String LOAD_TRAINING_CAZES = "Load training cases";
    private final static String LOAD_TEST_CAZES = "Load test cases";
    @FXML
    private TextField bitsPerWeight;
    @FXML
    private Button loadTrainingCazesButton;
    @FXML
    private Button loadTestCazesButton;
    @FXML
    private TextField crossoverPoints;
    @FXML
    private CheckBox uniformCrossover;
    @FXML
    private TextField mutationProbability;
    @FXML
    private CheckBox mutatePerComponent;
    @FXML
    private CheckBox usingModel2;
    @FXML
    private CheckBox usingSimpleKNN;
    @FXML
    private CheckBox onlyOtherLocations;
    @FXML
    private TextField kNN;
    @FXML
    private TextField numTestsPerIndividual;
    @FXML
    private TextField numTestsWhenFinished;
    private FileChooser fileChooser;
    private Set<Caze> trainingCazes = null;
    private Set<Caze> testCazes = null;

    public CbrProblem getProblemDefinition() throws UnexpectedException {
        boolean error = false;

        error |= trainingCazes == null;
        error |= !InputValidations.isInteger(crossoverPoints, 0, Integer.MAX_VALUE);
        error |= !InputValidations.isDouble(mutationProbability, 0., 1.);
        error |= !InputValidations.isInteger(numTestsPerIndividual, 0, Integer.MAX_VALUE);
        error |= !InputValidations.isInteger(bitsPerWeight, 1, Integer.MAX_VALUE);
        error |= !InputValidations.isInteger(kNN, 1, Integer.MAX_VALUE);
        if(testCazes != null) {
            error |= !InputValidations.isInteger(numTestsWhenFinished, 1, Integer.MAX_VALUE);
        }

        if(error) {
            throw new IllegalArgumentException("Illegal argument in CbrController");
        } else {

            return new CbrProblem(Integer.valueOf(crossoverPoints.getText()),
                                  uniformCrossover.isSelected(),
                                  mutatePerComponent.isSelected(),
                                  Double.valueOf(mutationProbability.getText()),
                                  Integer.valueOf(bitsPerWeight.getText()),
                                  usingModel2.isSelected(),
                                  trainingCazes,
                                  Integer.valueOf(numTestsPerIndividual.getText()),
                                  usingSimpleKNN.isSelected(),
                                  Integer.valueOf(kNN.getText()));
        }
    }

    @Override
    public void initialize(URL location, ResourceBundle resources) {
        fileChooser = new FileChooser();
        loadTrainingCazesButton.setText(LOAD_TRAINING_CAZES);
        loadTestCazesButton.setText(LOAD_TEST_CAZES);
        loadTrainingCazesButton.setOnAction(event -> loadCazes(event, LOAD_TRAINING_CAZES));
        loadTestCazesButton.setOnAction(event -> loadCazes(event, LOAD_TEST_CAZES));

        // Load case files if in workdir
        File trainingCazes = new File("trainingset.csv");
        File testCazes = new File("testset.csv");

        if(trainingCazes.exists() && trainingCazes.isFile()) {
            loadCazes(LOAD_TRAINING_CAZES, loadTrainingCazesButton, trainingCazes);
        }
        if(testCazes.exists() && testCazes.isFile()) {
            loadCazes(LOAD_TEST_CAZES, loadTestCazesButton, testCazes);
        }
    }

    void loadCazes(ActionEvent event, String cazeType) {
        Button sourceButton = (Button) event.getSource();
        fileChooser.setTitle("Select cases file");
        File file = fileChooser.showOpenDialog(loadTrainingCazesButton.getScene().getWindow());
        loadCazes(cazeType, sourceButton, file);
    }

    private void loadCazes(String cazeType, Button sourceButton, File file) {
        Set<Caze> cazes;
        if(file != null) {
            sourceButton.setText(cazeType + " [x]");
            try {
                cazes = CazeFileReader.readFile(file, Attributes.ALL_LOCATIONS);
            } catch(IOException | ParseException e) {
                throw new IllegalArgumentException("Could not parse the case file.", e);
            }
        } else {
            sourceButton.setText(cazeType + " [ ]");
            cazes = null;
        }

        switch(cazeType) {
            case LOAD_TRAINING_CAZES:
                trainingCazes = cazes;
                break;
            case LOAD_TEST_CAZES:
                testCazes = cazes;
                break;
        }
    }


    public void visualizeResults(CbrProblem problem, EvolutionaryAlgorithm.Results results) {
        if(Platform.isFxApplicationThread()) {
            _visualizeResults(problem, results);
        } else {
            Platform.runLater(() -> _visualizeResults(problem, results));
        }
    }

    private void _visualizeResults(CbrProblem problem, EvolutionaryAlgorithm.Results results) {
        CbrProblem.CbrPhenotype best = (CbrProblem.CbrPhenotype) results.bestIndividual;

        System.out.println("\n\n" + best.getInfo());

        if(testCazes != null) {
            System.out.println("\n\nStarting test with weights found.");
            Thread testThread = new Thread(() -> {
                try {
                    Logger logger = Logger.getLogger(this.getClass().getName());
                    if(usingModel2.isSelected()) {
                        Tests.runTests(logger, trainingCazes, testCazes, Integer.valueOf(kNN.getText()),
                                       Integer.valueOf(numTestsWhenFinished.getText()), onlyOtherLocations.isSelected(),
                                       new Model2(best.getWeights(), usingSimpleKNN.isSelected()));
                    } else {
                        Tests.runTests(logger, trainingCazes, testCazes, Integer.valueOf(kNN.getText()),
                                       Integer.valueOf(numTestsWhenFinished.getText()), onlyOtherLocations.isSelected(),
                                       new Model1(best.getWeights(), usingSimpleKNN.isSelected()));
                    }

                } catch(InterruptedException | ExecutionException e) {
                    e.printStackTrace();
                }
            });
            testThread.setDaemon(true);
            testThread.start();

            System.out.println(CbrProblem.explainWhetherSimpleKNN(usingSimpleKNN.isSelected()));
        }
    }

}
