package no.kolkinn.subsym.project2.genotypes;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.ThreadLocalRandom;

/**
 * Created by Mikael on 17.02.2016.
 */

public class BitVector {
    private final boolean[] vector;

    /**
     * Constructs a randomized bit vector with the given length.
     */
    public BitVector(int length) {
        vector = new boolean[length];

        for(int i = 0; i < vector.length; i++) {
            vector[i] = ThreadLocalRandom.current().nextBoolean();
        }
    }

    /**
     * Constructs a defensive copy of given BitVector
     */
    public BitVector(BitVector vector) {
        this.vector = vector.toBooleanArray();
    }

    /**
     * Constructs a BitVector from the given boolean array
     */
    public BitVector(boolean[] booleanArray) {
        this.vector = booleanArray.clone();
    }

    /**
     * @return defensive copy of boolean array representation
     */
    public boolean[] toBooleanArray() {
        return vector.clone();
    }

    protected void mutate(double probability, boolean perComponent) {
        if(perComponent) {
            for(int i = 0; i < vector.length; i++) {
                if(ThreadLocalRandom.current().nextDouble() < probability) {
                    vector[i] = !vector[i];
                }
            }
        } else if(ThreadLocalRandom.current().nextDouble() < probability) {
            int i = ThreadLocalRandom.current().nextInt(vector.length);
            vector[i] = !vector[i];
        }
    }

    protected BitVector[] crossover(BitVector mate, int nPoints, boolean uniform) {
        assert nPoints > 0 && nPoints < vector.length;

        boolean[] mateArray = mate.toBooleanArray();

        assert mateArray.length == vector.length;

        boolean[] childA;
        boolean[] childB;
        boolean temp;
        boolean selfToA;

        Set<Integer> crossoverPoints = new HashSet<>();
        while(crossoverPoints.size() < nPoints) {
            crossoverPoints.add(ThreadLocalRandom.current().nextInt(vector.length));
        }

        if(uniform) {
            childA = toBooleanArray();
            childB = mateArray;

            for(Integer i : crossoverPoints) {
                temp = childA[i];
                childA[i] = childB[i];
                childB[i] = temp;
            }

        } else {
            childA = new boolean[vector.length];
            childB = new boolean[vector.length];
            selfToA = ThreadLocalRandom.current().nextBoolean();

            for(int i = 0; i < vector.length; i++) {
                if(crossoverPoints.contains(i)) {
                    selfToA = !selfToA;
                }

                if(selfToA) {
                    childA[i] = vector[i];
                    childB[i] = mateArray[i];
                } else {
                    childA[i] = mateArray[i];
                    childB[i] = vector[i];
                }
            }
        }

        return new BitVector[] { new BitVector(childA), new BitVector(childB) };
    }

    @Override
    public int hashCode() {
        return Arrays.hashCode(vector);
    }

    @Override
    public boolean equals(Object obj) {
        if(obj instanceof BitVector) {
            return (Arrays.equals(vector, ((BitVector) obj).toBooleanArray()));
        }
        return false;
    }

    @Override
    public String toString() {
        StringBuilder builder = new StringBuilder();
        builder.append("[");
        for(int i = 0; i < vector.length; i++) {
            builder.append(vector[i] ? "1" : "0");
            if(i < vector.length - 1) {
                builder.append(", ");
            }
        }
        builder.append("]");
        return builder.toString();
    }
}
