package no.kolkinn.subsym.project2.problems;

import no.kolkinn.cbr_dovre.cbr.CBR;
import no.kolkinn.cbr_dovre.cbr.Caze;
import no.kolkinn.cbr_dovre.cbr.SimilarityFunction;
import no.kolkinn.cbr_dovre.methods.maintenance.RemoveNonCausalClosedOrConvoy;
import no.kolkinn.cbr_dovre.methods.selection.Grouping;
import no.kolkinn.cbr_dovre.models.Model1;
import no.kolkinn.cbr_dovre.models.Model2;
import no.kolkinn.cbr_dovre.tests.Tests;
import no.kolkinn.subsym.project2.PhenotypeDevelopment;
import no.kolkinn.subsym.project2.Problem;
import no.kolkinn.subsym.project2.genotypes.BitVector;
import no.kolkinn.subsym.project2.phenotypes.DoubleVector;

import java.time.temporal.ChronoUnit;
import java.time.temporal.TemporalUnit;
import java.util.*;
import java.util.concurrent.*;

/**
 * Created by Mikael on 14.09.2016.
 */
public class CbrProblem implements Problem {
    private final int nPoints;
    private final boolean uniform;
    private final boolean mutatePerComponent;
    private final double mutationProbability;
    private final int numBitsPerWeight;
    private final boolean usingModel2;
    private final Set<Caze> cazes;
    private final int numTestsPerIndividual;
    private final boolean simpleKNN;
    private final int kNN;
    private final ThreadPoolExecutor executor;
    private final TestScenario[] testScenarios;
    private final int numWeights;
    private volatile boolean stop;
    private boolean pruneTrainingCases = false;
    private long groupDistance = 24;
    private TemporalUnit unit = ChronoUnit.HOURS;

    /**
     * @param nPoints               Number of points used for crossover
     * @param uniform               If uniform, only the respective crossover points will be interchanged, not the gaps between them
     * @param mutatePerComponent    If true, each bit in the genotype is mutated with the probability _mutationProbability_, otherwise just one random bit may be mutated
     * @param mutationProbability   Probability that a mutation occurs
     * @param numBitsPerWeight      Number of bits used to represent each weight. End result (phenotype) is a number between and including 0.0 and 1.0
     * @param usingModel2           Use {@link Model2} if true, else {@link Model1}
     * @param cazes                 Cases to use both for training and testing
     * @param numTestsPerIndividual Number of different CBR case base and test set compositions to test per generation
     * @param simpleKNN             if true all cases will be treated similarly, otherwise the similarity measures will differ based on the result in the case a query is tested against
     * @param kNN                   number of nearest neighbours to use for making a decision
     */
    public CbrProblem(int nPoints, boolean uniform, boolean mutatePerComponent, double mutationProbability,
                      int numBitsPerWeight, boolean usingModel2, Set<Caze> cazes, int numTestsPerIndividual,
                      boolean simpleKNN, int kNN) {
        this.nPoints = nPoints;
        this.uniform = uniform;
        this.mutatePerComponent = mutatePerComponent;
        this.mutationProbability = mutationProbability;
        this.numBitsPerWeight = numBitsPerWeight;
        this.usingModel2 = usingModel2;
        this.cazes = cazes;
        this.numTestsPerIndividual = numTestsPerIndividual;
        this.simpleKNN = simpleKNN;
        this.kNN = kNN;

        numWeights = this.usingModel2 ? Model2.ATTRIBUTES.length : Model1.ATTRIBUTES.length;
        testScenarios = new TestScenario[numTestsPerIndividual];
        stop = false;

        Thread.UncaughtExceptionHandler exceptionHandler = (th, ex) -> {
            System.err.println("Uncaught exception in EA pool for CBR fitness evaluations: " + ex);
            ex.printStackTrace();
        };
        int numThreads = Runtime.getRuntime().availableProcessors();
//        if(numThreads > 1) {
//            numThreads -= numThreads * 0.1;
//        }
        executor = new ThreadPoolExecutor(numThreads,
                                          Runtime.getRuntime().availableProcessors(), 3,
                                          TimeUnit.SECONDS, new LinkedBlockingQueue<>(), r -> {
            Thread t = Executors.defaultThreadFactory().newThread(
                    r);
            t.setUncaughtExceptionHandler(exceptionHandler);
            t.setDaemon(true);
            return t;
        });
    }

    public static String explainWhetherSimpleKNN(boolean simpleKNN) {
        if(simpleKNN) {
            return "Same similarity measure regardless of case result.";
        } else {
            return "Using special similarity measures for Open and Closed.";
        }
    }

    @Override
    public Genotype generateGenotype() {
        return new CbrGenotype(numBitsPerWeight * numWeights);
    }

    @Override
    public void prepareNewGeneration(int generationNumber) {
        // Prepare a new set of case bases and test sets to be used each generation
        for(int i = 0; i < testScenarios.length; i++) {
            testScenarios[i] = new TestScenario(cazes);
        }
    }

    @Override
    public void stop() {
        stop = true;
        executor.shutdown();
    }

    /**
     * Prune cases with a phenotype's weights before testing
     * its fitness.
     */
    public void setPruneTrainingCases(boolean prune, long groupDistance, TemporalUnit timeUnit) {
        pruneTrainingCases = prune;
        this.groupDistance = groupDistance;
        this.unit = timeUnit;
    }

    private class TestScenario {
        Set<Caze> cazeBase;
        Set<Caze> testSet;

        private TestScenario(Set<Caze> cazes) {
            cazeBase = new HashSet<>();
            testSet = new HashSet<>();

            List<Set<Caze>> groups = new ArrayList<>(
                    Grouping.groupConnectedCases(cazes, groupDistance, unit));
            Collections.shuffle(groups, ThreadLocalRandom.current());

            int numTrainingGroupsGoal = (int) (0.70 * groups.size());
            for(int i = 0; i < groups.size(); i++) {
                if(i == 0) {
                    // Add at least one group to the training set
                    cazeBase.addAll(groups.get(i));
                } else if(i == groups.size() - 1) {
                    // Add at least one group to the test set
                    testSet.addAll(groups.get(i));
                } else if(i < numTrainingGroupsGoal) {
                    // i indicates the number of groups added to trainingSet
                    cazeBase.addAll(groups.get(i));
                } else {
                    testSet.addAll(groups.get(i));
                }
            }
        }
    }


    class CbrGenotype extends BitVector implements Problem.Genotype {


        public CbrGenotype(int length) {
            super(length);
        }

        public CbrGenotype(BitVector vector) {
            super(vector);
        }

        public CbrGenotype(boolean[] booleanArray) {
            super(booleanArray);
        }

        @Override
        public Genotype[] crossover(Genotype mate) {
            no.kolkinn.subsym.project2.genotypes.BitVector[] vectors = super.crossover(
                    (no.kolkinn.subsym.project2.genotypes.BitVector) mate, nPoints, uniform);
            Genotype[] children = new Genotype[vectors.length];
            for(int i = 0; i < vectors.length; i++) {
                children[i] = new CbrGenotype(vectors[i]);
            }
            return children;
        }

        @Override
        public void mutate() {
            super.mutate(mutationProbability, mutatePerComponent);
        }

        @Override
        public Genotype copy() {
            return new CbrGenotype(this);
        }

        @Override
        public Phenotype developPhenotype() {
            return new CbrPhenotype(this);
        }
    }

    public class CbrPhenotype extends DoubleVector implements Problem.Phenotype {
        private final CbrGenotype genotype;
        private final ArrayList<Future<Double>> fitnessFutures;
        double fitness;

        public CbrPhenotype(CbrGenotype genotype) {
            super(PhenotypeDevelopment.bitVectorToDoubleVector(genotype, numBitsPerWeight, 0., 1.));
            this.genotype = genotype;

            fitness = Double.NaN;
            fitnessFutures = new ArrayList<>();
            SimilarityFunction similarityFunction;
            if(usingModel2) {
                similarityFunction = new Model2(getWeights(), simpleKNN);
            } else {
                similarityFunction = new Model1(getWeights(), simpleKNN);
            }

            // Run
            for(TestScenario scenario : testScenarios) {
                fitnessFutures.add(executor.submit(new FitnessEvaluator(scenario, similarityFunction)));
            }
        }

        public Map<String, Double> getWeights() {
            Map<String, Double> weights = new HashMap<>();
            String[] attributes = usingModel2 ? Model2.ATTRIBUTES : Model1.ATTRIBUTES;

            for(int i = 0; i < attributes.length; i++) {
                weights.put(attributes[i], vector[i]);
            }

            return weights;
        }

        @Override
        public Genotype getGenotype() {
            return genotype;
        }

        @Override
        public double getFitness() {
            if(Double.isNaN(fitness)) {
                fitness = 0.0;
                try {
                    for(Future<Double> future : fitnessFutures) {
                        fitness += future.get();
                    }

                    fitness /= fitnessFutures.size();

                } catch(InterruptedException | ExecutionException e) {
                    throw new RuntimeException(e);
                }
            }

            return fitness;
        }

        @Override
        public String getInfo() {
            return toString();
        }

        @Override
        public String toString() {
            return getWeights().toString();
        }

        class FitnessEvaluator implements Callable<Double> {

            private final TestScenario scenario;
            private final SimilarityFunction similarityFunction;

            public FitnessEvaluator(TestScenario scenario, SimilarityFunction similarityFunction) {
                this.scenario = scenario;
                this.similarityFunction = similarityFunction;
            }

            @Override
            public Double call() throws Exception {
                Set<Caze> cazeBase = scenario.cazeBase;
                if(pruneTrainingCases) {
                    RemoveNonCausalClosedOrConvoy pruner = new RemoveNonCausalClosedOrConvoy(
                            new CBR(similarityFunction, kNN));
                    cazeBase = pruner.removeNonCausalClosedOrConvoy(cazeBase, groupDistance, unit);
                }

                return Tests.testCBR(null, new CBR(similarityFunction, kNN), cazeBase,
                                     null, Grouping.groupCasesByTime(scenario.testSet));
            }
        }
    }
}
