package no.kolkinn.subsym.project2.problems;

import no.kolkinn.subsym.project2.PhenotypeDevelopment;
import no.kolkinn.subsym.project2.Problem;
import no.kolkinn.subsym.project2.phenotypes.IntVector;

import java.util.*;

/**
 * Created by Mikael on 23.02.2016.
 */
public class Surprising implements Problem {
    public final int nPoints;
    public final boolean uniform;
    public final boolean mutatePerComponent;
    public final double mutationProbability;
    public final int symbolSize;
    public final int length;
    public final boolean global;
    public final boolean enableCache;
    protected final int bitsPerInt;
    protected Map<String, Double> genotypeFitness;

    public Surprising(int nPoints, boolean uniform, boolean mutatePerComponent, double mutationProbability, int symbolSize, int length, boolean global, boolean enableCache) {
        this.nPoints = nPoints;
        this.uniform = uniform;
        this.mutatePerComponent = mutatePerComponent;
        this.mutationProbability = mutationProbability;
        this.symbolSize = symbolSize;
        this.length = length;
        this.global = global;
        this.enableCache = enableCache;
        this.bitsPerInt = numBitsToRepresentSymbols(symbolSize);
        this.genotypeFitness = new HashMap<>();
    }

    public void setCache(Map<String, Double> cache) {
        genotypeFitness = cache;
    }

    public Map<String, Double> getCache() {
        return genotypeFitness;
    }

    @Override
    public Genotype generateGenotype() {
        return new SurprisingGenotype(symbolSize, length);
    }

    @Override
    public void prepareNewGeneration(int generationNumber) {
    }

    @Override
    public void stop() {

    }

    public static int numBitsToRepresentSymbols(int symbolSize) {
        int numBits = 1;

        while((symbolSize >>= 1) != 0) {
            numBits++;
        }
        return numBits;
    }

    class SurprisingGenotype extends no.kolkinn.subsym.project2.genotypes.IntVector implements Genotype {
        public SurprisingGenotype(int symbolSize, int length) {
            super(symbolSize, length);
        }

        public SurprisingGenotype(no.kolkinn.subsym.project2.genotypes.IntVector vector) {
            super(vector);
        }

        public SurprisingGenotype(int[] intArray, int symbolSize) {
            super(intArray, symbolSize);
        }

        @Override
        public Genotype[] crossover(Genotype mate) {
            no.kolkinn.subsym.project2.genotypes.IntVector[] vectors = super.crossover((no.kolkinn.subsym.project2.genotypes.IntVector) mate, nPoints, uniform);
            Genotype[] children = new Genotype[vectors.length];
            for(int i = 0; i < vectors.length; i++) {
                children[i] = new SurprisingGenotype(vectors[i]);
            }
            return children;
        }

        @Override
        public void mutate() {
            mutate(mutationProbability, mutatePerComponent);
        }

        @Override
        public Genotype copy() {
            return new SurprisingGenotype(this);
        }

        @Override
        public Phenotype developPhenotype() {
            return new SurprisingPhenotype(this);
        }
    }

    class SurprisingPhenotype extends IntVector implements Phenotype {
        protected final SurprisingGenotype genotype;  // TODO: 20.02.2016 Make immutable?
        protected double fitness = Double.NaN;

        public SurprisingPhenotype(SurprisingGenotype genotype) {
            super(PhenotypeDevelopment.intVectorToIntVector(genotype));
            this.genotype = genotype;
        }

        @Override
        public Genotype getGenotype() {
            return genotype;
        }

        @Override
        public synchronized double getFitness() {
            // Only calculate once
            if(Double.isNaN(fitness)) {
                if(enableCache && genotypeFitness.containsKey(genotype.hashCode() + "")) {
                    fitness = genotypeFitness.get(genotype.hashCode() + "");
                } else {
                    Set<Integer> patterns = new HashSet<>();
                    int conflicts = 0;

                    for(int i = 0; i < vector.length - 1; i++) {
                        // If global, test all distances. If local, fixed distance of 0
                        for(int j = i + 1; j < (global ? vector.length : i + 2); j++) {
                            if(!patterns.add(Arrays.hashCode(new int[] { vector[i], vector[j], j - i - 1 }))) {
                                conflicts++;
                            }
                        }
                    }
                    fitness = 1.0 / (conflicts + 1);

                    if(enableCache) {
                        genotypeFitness.put(genotype.hashCode() + "", fitness);
                    }
                }
            }
            return fitness;
        }

        @Override
        public String getInfo() {
            return String.format("Symbol size: %3d \t Length: %3d \t Globally surprising: %b \t N-point: %3d \t Uniform: %b \t Mutate per: %b \t Mutation probability: %1.5f",
                                 symbolSize, length, global, nPoints, uniform, mutatePerComponent, mutationProbability);
        }
    }
}
