package no.kolkinn.cbr_dovre.cbr;

import java.io.IOException;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.concurrent.PriorityBlockingQueue;
import java.util.function.BiConsumer;
import java.util.logging.FileHandler;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collector;

public class CBR {

    private final SimilarityFunction similarityFunction;
    private final int kNN;
    private Set<Caze> cazeBase = new HashSet<>();

    public CBR(SimilarityFunction similarityFunction, int kNN) {
        this.similarityFunction = similarityFunction;
        this.kNN = kNN;
    }

    public static <T> Collector<T, ?, List<T>> maxN(Comparator<? super T> comparator, int limit) {
        BiConsumer<PriorityBlockingQueue<T>, T> accumulator = (queue, t) -> {
            queue.add(t);
            if(queue.size() > limit)
                queue.poll();
        };
        return Collector.of(() -> new PriorityBlockingQueue<>(limit + 1, comparator),
                            accumulator, (q1, q2) -> {
                    for(T t : q2) {
                        accumulator.accept(q1, t);
                    }
                    return q1;
                }, ArrayList::new);
    }

    public void setCazeBase(Set<Caze> cazeBase) {
        this.cazeBase = cazeBase;
    }

    public Set<Caze> getCazeBase() {
        return cazeBase;
    }

    public SimilarityFunction getSimilarityFunction() {
        return similarityFunction;
    }

    public Decision makeDecision(Caze query) {
        List<CaseSimilarity> nearest = mostSimilarCazes(query, kNN);

        Decision bestDecision = null;
        boolean sorted = false;

        for(int i = kNN; i > 0; i--) {
            if(i != kNN && nearest.size() > 1) {
                // Shrink the list by one, removing the least similar neighbour
                // Don't shrink the first round

                if(!sorted) {
                    // The k nearest neighbours are not sorted by default.
                    // Since k is usually a small number, sorting is cheap
                    // and makes removing the least close neighbour easy.
                    nearest.sort(Comparator.comparingDouble(caseSimilarity -> caseSimilarity.similarity));
                    sorted = true;
                }

                nearest = nearest.subList(1, nearest.size());
            }

            if(nearest.size() == 1) {
                return nearest.get(0).caze.getDecision();
            } else if(nearest.size() < 1) {
                return null;
            }

            Map<Decision, Double> decisionSum = new HashMap<>(Decision.values().length);
            for(CaseSimilarity caseSimilarity : nearest) {
                decisionSum.compute(caseSimilarity.caze.getDecision(), (d, sum) -> (sum == null) ? caseSimilarity.similarity : sum + caseSimilarity.similarity);
            }

            bestDecision = null;
            double bestSum = 0.0;
            boolean unanimous = false;
            for(Decision decision : Decision.values()) {
                double tempSum = decisionSum.getOrDefault(decision, 0.0);
                if(tempSum > bestSum) {
                    bestSum = tempSum;
                    bestDecision = decision;
                    unanimous = true;
                } else if(tempSum == bestSum) {
                    unanimous = false;
                }
            }
            if(unanimous) {
                // A unanimous decision has been made
                break;
            }
        }

        return bestDecision;
    }

    public List<CaseSimilarity> mostSimilarCazes(Caze query, int limit) {
        try {
            return cazeBase.stream().map(caze -> new CaseSimilarity(caze, query)).collect(
                    maxN(Comparator.comparingDouble(caseSimilarity -> caseSimilarity.similarity), limit));
        } catch(Exception e) {
            LocalDateTime dateTime = LocalDateTime.now();
            String logDate = dateTime.format(DateTimeFormatter.ofPattern("dd-MM-uuuu_HH-mm", Locale.getDefault()));
            Logger logger = Logger.getAnonymousLogger(); // Make each logger unique to the instance
            try {
                logger.addHandler(new FileHandler("./configuration_test_" + logDate + ".%u.log"));
                logger.log(Level.SEVERE, "Exception in mostSimilarCazes():", e);
                logger.log(Level.SEVERE, "Cause for exception:", e.getCause());
                logger.severe(String.format("cazeBase.size(): %d\n"
                                            + "query: %s\n"
                                            + "cazeBase: %s\n",
                                            cazeBase == null ? -1 : cazeBase.size(),
                                            query.toString(),
                                            cazeBase == null ? "null" : Arrays.toString(cazeBase.toArray())));

            } catch(IOException e1) {
                e.printStackTrace();
                e1.printStackTrace();
            }

            throw new RuntimeException(e);
        }
    }

    public Caze mostSimilarCaze(Caze query) {
        Caze bestCaze = null;
        double bestSimilarity = Double.NEGATIVE_INFINITY;
        for(Caze caze : cazeBase) {
            double similarity = similarityFunction.calculateSimilarity(caze, query);
            if(similarity > bestSimilarity) {
                bestCaze = caze;
                bestSimilarity = similarity;
            }
        }
        return bestCaze;
    }

    class CaseSimilarity {
        Caze query, caze;
        double similarity;

        public CaseSimilarity(Caze caze, Caze query) {
            this.query = query;
            this.caze = caze;
            this.similarity = similarityFunction.calculateSimilarity(caze, query);
        }

        public CaseSimilarity(Caze caze, Caze query, double similarity) {
            this.query = query;
            this.caze = caze;
            this.similarity = similarity;
        }
    }
}
