package no.kolkinn.cbr_dovre.cbr;

import com.sun.istack.internal.NotNull;
import no.kolkinn.cbr_dovre.domain.SpeedLimitCalculation;
import no.kolkinn.cbr_dovre.domain.VisibilityEstimation;
import no.kolkinn.cbr_dovre.models.Attributes;

import java.time.LocalDateTime;
import java.util.*;
import java.util.Map.Entry;

public class Caze {

    public static final Random RANDOM = new Random();

    private final LocalDateTime timestamp;
    private final Map<String, Double> values;
    private final Map<String, Double> cache;
    private final Decision decision;
    private final String location;

    public Caze(@NotNull LocalDateTime timestamp, @NotNull String location, Map<String, Double> values,
                Decision decision) {
        if(timestamp == null) {
            throw new IllegalArgumentException("timestamp cannot be null!");
        }
        if(location == null) {
            throw new IllegalArgumentException("location cannot be null!");
        }

        this.timestamp = timestamp;
        this.location = location;
        this.values = Collections.unmodifiableMap(new HashMap<>(values));
        this.cache = new HashMap<>();
        this.decision = decision;
    }

    public LocalDateTime getTimestamp() {
        return timestamp;
    }

    public Decision getDecision() {
        return decision;
    }

    public String getLocation() {
        return location;
    }

    public Double getAttributeValue(String attribute) {
        if(values.containsKey(attribute)) {
            return values.get(attribute);
        } else if(cache.containsKey(attribute)) {
            return cache.get(attribute);
        } else if(attribute.startsWith(Attributes.VISIBILITY)) {
            Double precipitation = getAttributeValue(Attributes.PRECIPITATION);
            Double temperature = getAttributeValue(Attributes.TEMPERATURE);
            Double windSpeed = getAttributeValue(Attributes.WIND_SPEED);
            Double visibility = null;
            if(precipitation != null && temperature != null && windSpeed != null) {
                visibility = VisibilityEstimation.visibility(precipitation, temperature, windSpeed);
            }
            cache.put(attribute, visibility);
            return visibility;
        } else if(attribute.startsWith(Attributes.SPEED_LIMIT)) {
            Double windSpeed = getAttributeValue(Attributes.WIND_SPEED);
            Double friction = getAttributeValue(Attributes.FRICTION);
            Double visibility = getAttributeValue(Attributes.VISIBILITY);
            Double speedLimit = null;
            if(windSpeed != null && friction != null && visibility != null) {
                speedLimit = SpeedLimitCalculation.speedLimit(windSpeed, friction, visibility);
            }
            cache.put(attribute, speedLimit);
            return speedLimit;
        } else {
            return null;
        }
    }

    public Set<String> getAttributes() {
        Set<String> attributes = new HashSet<>(values.keySet());

        // Virtual attributes
        attributes.add(Attributes.VISIBILITY);
        attributes.add(Attributes.SPEED_LIMIT);

        return attributes;
    }

    public Caze applySensorUncertainty() {
        Map<String, Double> newValues = new HashMap<>();
        for(Entry<String, Double> entry : values.entrySet()) {
            String attribute = entry.getKey();
            Double value = entry.getValue();
            if(attribute.startsWith(Attributes.WIND_SPEED)) {
                Double newValue = null;
                if(value != null) {
                    if(value < 5.0) {
                        newValue = value + (RANDOM.nextDouble() - 0.5);
                        newValue = Math.max(newValue, 0.0);
                    } else {
                        newValue = value * (0.2 * RANDOM.nextDouble() + 0.9);
                        newValue = Math.max(newValue, 0.0);
                    }
                }
                newValues.put(attribute, newValue);
            } else if(attribute.startsWith(Attributes.WIND_DIRECTION)) {
                Double newValue = null;
                if(value != null) {
                    newValue = value + (10.0 * RANDOM.nextDouble() - 5.0);
                    if(newValue < 0.0) {
                        newValue += 360.0;
                    }
                    if(newValue >= 360.0) {
                        newValue -= 360.0;
                    }
                }
                newValues.put(attribute, newValue);
            } else if(attribute.startsWith(Attributes.FRICTION)) {
                newValues.put(attribute, value);
            } else if(attribute.startsWith(Attributes.FRICTION)) {
                Double newValue = null;
                if(value != null) {
                    if(value >= -10.0 && value <= 10.0) {
                        newValue = value + (0.2 * RANDOM.nextDouble() - 0.1);
                    } else {
                        newValue = value + (0.4 * RANDOM.nextDouble() - 0.2);
                    }
                }
                newValues.put(attribute, newValue);
            } else if(attribute.startsWith(Attributes.PRECIPITATION)) {
                Double newValue = null;
                if(value != null) {
                    newValue = value + (0.2 * RANDOM.nextDouble() - 0.1);
                    newValue = Math.max(newValue, 0.0);
                }
                newValues.put(attribute, newValue);
            } else {
                newValues.put(attribute, value);
            }
        }
        return new Caze(timestamp, location, newValues, decision);
    }

    @Override
    public boolean equals(Object o) {
        if(this == o) return true;
        if(!(o instanceof Caze)) return false;

        Caze caze = (Caze) o;

        if(!timestamp.equals(caze.timestamp)) return false;
        if(!values.equals(caze.values)) return false;
        if(decision != caze.decision) return false;
        return location.equals(caze.location);

    }

    @Override
    public int hashCode() {
        int result = timestamp.hashCode();
        result = 31 * result + values.hashCode();
        result = 31 * result + (decision != null ? decision.hashCode() : 0);
        result = 31 * result + location.hashCode();
        return result;
    }
}
