package no.kolkinn.cbr_dovre.domain;

import no.kolkinn.cbr_dovre.cbr.Decision;

public class SpeedLimitCalculation {

    private static final double REACTION_TIME = 1.0;
    private static final double GRAVITY = 9.81;
    private static final double VISIBILITY_SAFETY_DISTANCE = 10.0;
    private static final double LOWER_SPEED_LIMIT_CONVOY = 25.5;
    private static final double LOWER_SPEED_LIMIT_OPEN = 40.5;

    public static double speedLimit(double windSpeed, double friction, double visibility) {
        double speedLimit = Double.POSITIVE_INFINITY;
        speedLimit = Math.min(speedLimit, speedLimitVisibilityFriction(friction, visibility));
        speedLimit = Math.min(speedLimit, speedLimitWindFrictionBus(windSpeed, friction));
        speedLimit = Math.min(speedLimit, speedLimitWindFrictionDoubleDecker(windSpeed, friction));
        return speedLimit;
    }

    public static Decision decision(double windSpeed, double friction, double visibility) {
        double speedLimit = speedLimit(windSpeed, friction, visibility);
        if(speedLimit < LOWER_SPEED_LIMIT_CONVOY) {
            return Decision.CLOSED;
        } else if(speedLimit < LOWER_SPEED_LIMIT_OPEN) {
            return Decision.CONVOY;
        } else {
            return Decision.OPEN;
        }
    }

    private static double speedLimitVisibilityFriction(double friction, double visibility) {
        double speedLimit =  3.6 * friction * GRAVITY
               * (Math.sqrt(
                Math.pow(REACTION_TIME, 2) + (visibility - VISIBILITY_SAFETY_DISTANCE) / (friction * GRAVITY))
                  - REACTION_TIME);
        return Double.isNaN(speedLimit) ? 0.0 : speedLimit;
    }

    private static double speedLimitWindFrictionBus(double windSpeed, double friction) {
        int windSpeedRange = windSpeedRange(windSpeed);
        if(windSpeedRange != -1) {
            int frictionRange = frictionRangeBus(windSpeedRange, friction);
            if(frictionRange != -1) {
                double[] speedLimits = { 80, 70, 60, 50, 40, 30 };
                return speedLimits[frictionRange];
            }
        }
        return 0;
    }

    private static double speedLimitWindFrictionDoubleDecker(double windSpeed, double friction) {
        int windSpeedRange = windSpeedRange(windSpeed);
        if(windSpeedRange != -1) {
            int frictionRange = frictionRangeDoubleDecker(windSpeedRange, friction);
            if(frictionRange != -1) {
                double[] speedLimits = { 80, 70, 60, 50, 40, 30 };
                return speedLimits[frictionRange];
            }
        }
        return 0;
    }

    private static int windSpeedRange(double windSpeed) {
        double[] windSpeedLimits = { 8.3, 10.0, 11.7, 13.3, 15.0, 16.7, 18.3, 20.0, 21.7, 23.3 };
        for(int i = 0; i < windSpeedLimits.length; i++) {
            if(windSpeed <= windSpeedLimits[i]) {
                return i;
            }
        }
        return -1;
    }

    private static int frictionRangeBus(int windSpeedRange, double friction) {
        double[][] frictionLimits = { { 0.100, 0.070, 0.065, 0.060, 0.058, 0.054 },
                                      { 0.120, 0.090, 0.080, 0.070, 0.060, 0.050 }, { 0.145, 0.115, 0.101, 0.086, 0.073, 0.061 },
                                      { 0.170, 0.140, 0.120, 0.100, 0.085, 0.070 }, { 0.210, 0.170, 0.150, 0.125, 0.113, 0.085 },
                                      { 0.255, 0.205, 0.180, 0.155, 0.140, 0.105 }, { 0.300, 0.245, 0.218, 0.190, 0.170, 0.135 },
                                      { 0.360, 0.290, 0.263, 0.235, 0.210, 0.180 }, { 0.420, 0.340, 0.310, 0.280, 0.250, 0.220 },
                                      { 0.500, 0.400, 0.370, 0.340, 0.310, 0.280 } };
        for(int i = 0; i < frictionLimits[windSpeedRange].length; i++) {
            if(friction >= frictionLimits[windSpeedRange][i]) {
                return i;
            }
        }
        return -1;
    }

    private static int frictionRangeDoubleDecker(int windSpeedRange, double friction) {
        double[][] frictionLimits = { { 0.130, 0.100, 0.095, 0.090, 0.088, 0.084 },
                                      { 0.160, 0.130, 0.120, 0.110, 0.100, 0.090 }, { 0.192, 0.162, 0.148, 0.133, 0.120, 0.108 },
                                      { 0.230, 0.200, 0.180, 0.160, 0.145, 0.130 }, { 0.275, 0.235, 0.215, 0.190, 0.178, 0.150 },
                                      { 0.325, 0.275, 0.250, 0.225, 0.210, 0.175 }, { 0.370, 0.315, 0.288, 0.260, 0.240, 0.205 },
                                      { 0.430, 0.355, 0.328, 0.300, 0.275, 0.245 }, { 0.490, 0.410, 0.380, 0.350, 0.320, 0.290 },
                                      { 0.550, 0.450, 0.420, 0.390, 0.360, 0.330 } };
        for(int i = 0; i < frictionLimits[windSpeedRange].length; i++) {
            if(friction >= frictionLimits[windSpeedRange][i]) {
                return i;
            }
        }
        return -1;
    }
}
