package no.kolkinn.cbr_dovre.methods.maintenance;

import no.kolkinn.cbr_dovre.cbr.CBR;
import no.kolkinn.cbr_dovre.cbr.Caze;
import no.kolkinn.cbr_dovre.cbr.SimilarityFunction;

import java.util.*;
import java.util.function.Consumer;

/**
 * Created by Mikael on 15.11.2016.
 */
public class ICFRedundancyRemoval {

    private final SimilarityFunction similarityFunction;

    public ICFRedundancyRemoval(CBR cbr) {
        similarityFunction = cbr.getSimilarityFunction();
    }

    public ICFRedundancyRemoval(SimilarityFunction similarityFunction) {
        this.similarityFunction = similarityFunction;
    }

    /**
     * Iterative Case Filtering algorithm (ICF)
     * <p>
     * ICF removes cases that are solved by more cases than
     * they themselves solve.
     *
     * @param cases cases to perform redundancy removal on (the collection is not modified)
     * @return a set containing the remaining cases after redundancy removal
     */
    public Set<Caze> removeDuplicates(Collection<Caze> cases) {
        /* ICF Algorithm:
         * T: Training Set
		 *
		 * Run RENN on T
		 * (Not included here, RENN performed seperately)
		 *
		 * Repeat
		 * 		For all x E T do
		 * 			compute reachable(x)
		 * 			compute coverage(x)
		 * 		End-For
		 * 		progress = false
		 * 		For all x E T do
		 * 			If |reachable(x)| > |coverage(x)| then
		 * 				flag x for removal
		 * 				process = true
		 * 			End-If
		 * 		End-For
		 * 		For all x E T do
		 * 			If x flagged for removal then
		 * 				T = T - {x}
		 * 			End-If
		 * 		End-For
		 * Until not progress
		 *
		 * Return T
		 */

        Consumer<Integer> logger = Log.getLogger("ICF");

        Set<Caze> remainingCases = new HashSet<>(cases);

        Map<Caze, Collection<Caze>> reachable = new HashMap<>();
        Map<Caze, Collection<Caze>> coverage = new HashMap<>();
        boolean changes;
        do {
            changes = false;

            reachable.clear();
            coverage.clear();
            computeReachabilityAndCoverageSets(reachable, coverage, remainingCases);

            Set<Caze> casesToRemove = new HashSet<>();
            for(Caze caze : remainingCases) {
                if(reachable.get(caze).size() > coverage.get(caze).size()) {
                    casesToRemove.add(caze);
                    changes = true;
                }
            }

            remainingCases.removeAll(casesToRemove);

            logger.accept(casesToRemove.size());

        } while(changes);

        return remainingCases;
    }

    /**
     * Computes the reachability and coverage sets for all cases
     * in the provided {@code caseBase} and put them in
     * {@code reachabilitySets} and {@code coverageSets} respectively.
     */
    private void computeReachabilityAndCoverageSets(Map<Caze, Collection<Caze>> reachabilitySets,
                                                    Map<Caze, Collection<Caze>> coverageSets,
                                                    Collection<Caze> caseBase) {
        if(caseBase == null) {
            return;
        }

        for(Caze query : new HashSet<>(caseBase)) {
            Set<Caze> solvesQuery = findCasesThatSolveQuery(query, caseBase);
            Collection<Caze> reachabilitySet = new HashSet<>();

            for(Caze caze : solvesQuery) {
                reachabilitySet.add(caze);

                Collection<Caze> coverageSet = coverageSets.getOrDefault(caze, new HashSet<>());
                coverageSet.add(query);
                coverageSets.put(caze, coverageSet);
            }

            reachabilitySets.put(query, reachabilitySet);
            coverageSets.putIfAbsent(query, new HashSet<>());   // Prevent null pointers by using an empty set
        }
    }

    /**
     * Finds the cases that solves a query.
     * <p>
     * In ICF the number of nearest neighbours varies,
     * and the neighbourhood is instead bounded by the
     * first case with a different decision, when starting
     * with the case with the highest similarity.
     */
    private Set<Caze> findCasesThatSolveQuery(Caze query, Collection<Caze> caseBase) {
        List<Caze> caseList = new ArrayList<>(caseBase);

        // Sort caseList with most similar first and least similar last
        caseList.sort(Comparator.comparingDouble(
                caze -> similarityFunction.calculateSimilarity((Caze) caze, query)).reversed());

        Set<Caze> casesThatSolveQuery = new HashSet<>();
        for(Caze caze : caseList) {
            if(caze.getDecision() != query.getDecision()) {
                break;
            }
            casesThatSolveQuery.add(caze);
        }

        return casesThatSolveQuery;
    }
}
