package no.kolkinn.cbr_dovre.methods.maintenance;

import no.kolkinn.cbr_dovre.cbr.CBR;
import no.kolkinn.cbr_dovre.cbr.Caze;
import no.kolkinn.cbr_dovre.cbr.Decision;
import no.kolkinn.cbr_dovre.methods.selection.Grouping;

import java.time.temporal.TemporalUnit;
import java.util.*;
import java.util.function.Consumer;

/**
 * Created by Mikael on 18.10.2016.
 */
public class RemoveNonCausalClosedOrConvoy {

    private final CBR cbr;

    public RemoveNonCausalClosedOrConvoy(CBR cbr) {
        this.cbr = cbr;
    }

    /**
     * Removes cases that likely did not cause the road to be closed or have convoys. This is relevant when the
     * the cases are from separate weather stations, but their stored decisions were taken on the basis of all
     * weather stations and possibly other information as well. When the road is closed because
     * of bad weather in one area, it doesn't necessarily mean that the weather dictates closing the road in
     * other parts of the same road.
     * <p/>
     * Works by grouping together cases based on their location and their timewise proximity, see {@link Grouping#groupConnectedCases(Set, long, TemporalUnit)}.
     * Each CLOSED and CONVOY case in a group is then queried against a {@link CBR} with the other groups as case base.
     * If the predicted decision is OPEN, the case is removed.
     * This procedure is repeated for each group.
     * <p/>
     * Assumes some differences between locations. If two locations, A and A', have basically the same values,
     * then a case in A is likely to have the "same case" in A' and the timewise surrounding cases of A' as
     * it's nearest neighbours.
     * This again means that a closed case where a location probably was not the cause of the closing,
     * is likely not to get removed because it's also non-causal relative will be it's closest match.
     * <p/>
     * If {@code cazes} is empty, an empty {@link Set} is returned
     *
     * @param cazes         cases to perform pruning on
     * @param groupDistance the minimum amount of time without any case in between two cases for them to be considered parts of separate groups
     * @param unit          the time unit used for {@code groupDistance}
     * @return the given set of {@code cazes} without the removed cases
     * @throws NullPointerException if {@code cazes} or {@code unit} is null, or if a case in {@code cazes} is null
     * @see Grouping#groupConnectedCases(Set, long, TemporalUnit) Grouping.groupConnectedCases used for grouping cases
     */
    public Set<Caze> removeNonCausalClosedOrConvoy(Collection<Caze> cazes, long groupDistance,
                                                   TemporalUnit unit) {
        if(cazes == null) {
            throw new NullPointerException("cazes cannot be null");
        }
        if(unit == null) {
            throw new NullPointerException("unit cannot be null");
        }
        if(cazes.size() == 0) {
            return new HashSet<>();
        }

        Set<Caze> currentCases = new HashSet<>(cazes);
        Set<Caze> removedCazes = new HashSet<>();

        Consumer<Integer> logger = Log.getLogger("NonCausal");
        int generation = 0;

        do {
            removedCazes.clear();

            // Group cases by location and timewise proximity
            List<Set<Caze>> groups = new ArrayList<>(Grouping.groupConnectedCases(currentCases, groupDistance, unit));

            for(int i = 0; i < groups.size(); i++) {
                Set<Caze> group = groups.get(i);

                // Use all other groups as case base
                Set<Caze> cazeBase = new HashSet<>();
                for(int j = 0; j < groups.size(); j++) {
                    if(j != i) {
                        cazeBase.addAll(groups.get(j));
                    }
                }
                cbr.setCazeBase(cazeBase);

                for(Caze caze : group) {

                    // Only work on convoy or closed cases
                    if(caze.getDecision() == Decision.CLOSED || caze.getDecision() == Decision.CONVOY) {
                        Decision decision = cbr.makeDecision(caze);

                        // If open, it very likely was not the cause of the closing or convoy,
                        // therefore we remove it
                        if(decision == Decision.OPEN) {
                            removedCazes.add(caze);
                        }
                    }
                }
            }

            currentCases.removeAll(removedCazes);
            logger.accept(removedCazes.size());

        } while(removedCazes.size() > 0 && ++generation < 1);

        return currentCases;
    }
}
