package no.kolkinn.cbr_dovre.methods.selection;

import no.kolkinn.cbr_dovre.cbr.Caze;

import java.time.LocalDateTime;
import java.time.temporal.TemporalUnit;
import java.util.*;
import java.util.stream.Collectors;

/**
 * Created by Mikael on 18.10.2016.
 */
public class Grouping {

    /**
     * Groups together cases from the same location that are closer to each other than {@code groupDistance}.
     * For a case to be considered a part the of a group, it must be closer than {@code groupDistance} to at
     * least one case in the group.
     * <p/>
     * If {@code cazes} is empty, an empty {@link Set} is returned
     *
     * @param cazes         cases to be grouped
     * @param groupDistance the minimum amount of time without any case in between two cases for them to be considered parts of separate groups
     * @param unit          the time unit used for {@code groupDistance}
     * @return a collection of case groups
     * @throws NullPointerException if {@code cazes} or {@code unit} is null, or if a case in {@code cazes} is null
     */
    public static Set<Set<Caze>> groupConnectedCases(Set<Caze> cazes, long groupDistance, TemporalUnit unit) {
        if(cazes == null) {
            throw new NullPointerException("cazes cannot be null");
        }
        if(unit == null) {
            throw new NullPointerException("unit cannot be null");
        }
        if(cazes.size() == 0) {
            return new HashSet<>();
        }

        // Sort the cases by time, from old to new
        Comparator<Caze> oldestFirst = Comparator.comparing(Caze::getTimestamp);

        LinkedList<Caze> sortedCazes = cazes.stream()
                                            .sorted(oldestFirst)
                                            .collect(Collectors.groupingBy(Caze::getLocation,
                                                                           LinkedHashMap::new, Collectors.toList()))
                                            .values().stream()
                                            .flatMap(Collection::stream)
                                            .collect(Collectors.toCollection(LinkedList::new));

        Set<Set<Caze>> groups = new HashSet<>();

        Set<Caze> group = null;
        String location = null;
        LocalDateTime latest = null;
        while(true) {
            if(sortedCazes.size() == 0) {
                // Save the last group
                groups.add(group);
                break;
            }

            Caze caze = sortedCazes.poll();
            if(caze == null) {
                throw new NullPointerException("Case in the sortedCazes list was null, but the list was not empty.");
            }

            LocalDateTime cazeTime = caze.getTimestamp();

            if(location == null || !location.equals(caze.getLocation())) {
                if(group != null) {
                    // Finished with the current location, save group and start the next one
                    groups.add(group);
                }

                group = new HashSet<>();
                location = caze.getLocation();
                latest = cazeTime;
                group.add(caze);
                continue;
            }

            // TODO (maybe): sanity check: see if cases are grouped together by location

            // Sanity check: It is required that the cases are sorted from old to new.
            if(cazeTime.isBefore(latest)) {
                throw new RuntimeException("sortedCazes list was not properly sorted");
            }

            long distance = latest.until(cazeTime, unit);
            if(distance < groupDistance) {
                // Case is within the groupDistance, add to group
                latest = cazeTime;
                group.add(caze);
            } else {
                // Distance is >= groupDistance, create a new group
                groups.add(group);
                group = new HashSet<>();
                latest = cazeTime;
                group.add(caze);
            }
        }

        return groups;
    }

    /**
     * Groups together cases from the same the same time. This would normally be sets of cases from the different locations
     * that had data at a given time.
     * <p/>
     * If {@code cazes} is empty, an empty {@link Set} is returned
     *
     * @param cazes cases to be grouped
     * @return a collection of case groups
     * @throws NullPointerException if {@code cazes} is null or if a case in {@code cazes} is null
     */
    public static Set<Set<Caze>> groupCasesByTime(Set<Caze> cazes) {
        if(cazes == null) {
            throw new NullPointerException("cazes cannot be null");
        }
        if(cazes.size() == 0) {
            return new HashSet<>();
        }

        // Sort the cases by time, from old to new
        Comparator<Caze> oldestFirst = Comparator.comparing(Caze::getTimestamp);

        LinkedList<Caze> sortedCazes = cazes.stream()
                                            .sorted(oldestFirst)
                                            .collect(Collectors.toCollection(LinkedList::new));

        Set<Set<Caze>> groups = new HashSet<>();

        Set<Caze> group = null;
        LocalDateTime current = null;
        while(true) {
            if(sortedCazes.size() == 0) {
                // Save the last group
                groups.add(group);
                break;
            }

            Caze caze = sortedCazes.poll();
            if(caze == null) {
                throw new NullPointerException("Case in the sortedCazes list was null, but the list was not empty.");
            }

            LocalDateTime cazeTime = caze.getTimestamp();

            // Sanity check: It is required that the cases are sorted from old to new.
            // Otherwise
            if(current != null && cazeTime.isBefore(current)) {
                throw new RuntimeException("sortedCazes list was not properly sorted.");
            }

            if(current == null || !current.equals(caze.getTimestamp())) {
                if(group != null) {
                    // Finished with the current location, save group and start the next one
                    groups.add(group);
                }

                group = new HashSet<>();
                current = cazeTime;
                group.add(caze);
                continue;
            }

            group.add(caze);
        }

        return groups;
    }
}
