package no.kolkinn.cbr_dovre.models;

import no.kolkinn.cbr_dovre.cbr.Caze;
import no.kolkinn.cbr_dovre.cbr.Decision;
import no.kolkinn.cbr_dovre.cbr.SimilarityFunction;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

public class SimilarityFunctions implements SimilarityFunction {

    private static final Map<String, Double> ROAD_DIRECTIONS;

    static {
        ROAD_DIRECTIONS = new HashMap<>();
//        ROAD_DIRECTIONS.put(Attributes.AVSJOEN, 60.0);
        ROAD_DIRECTIONS.put(Attributes.FOKSTUGU, 45.0);
        ROAD_DIRECTIONS.put(Attributes.HJERKINN, 0.0);
    }

    final String[] attributes;
    final boolean simpleKNN;
    Map<String, Double> weights;

    /**
     * @param attributes
     * @param weights
     * @param simpleKNN  if true all cases will be treated similarly, otherwise the similarity measures will differ based on the result in the case a query is tested against
     */
    public SimilarityFunctions(String[] attributes, Map<String, Double> weights, boolean simpleKNN) {
        this.attributes = attributes;
        this.weights = weights;
        this.simpleKNN = simpleKNN;
    }

    /**
     * Creates a copy of the original
     */
    public SimilarityFunctions(SimilarityFunctions original) {
        this(Arrays.copyOf(original.attributes, original.attributes.length), new HashMap<>(original.weights), original.simpleKNN);
    }

    /**
     * Takes the default weights and replaces them if other weights are supplied.
     * Default weights will remain for attributes that are not set.
     */
    protected static Map<String, Double> replaceWeights(String[] attributes, Map<String, Double> defaultWeights,
                                                        Map<String, Double> weights) {
        Map<String, Double> resultingWeights = new HashMap<>(defaultWeights);
        Double weight;

        if(weights != null) {
            for(String attribute : attributes) {
                if(weights.containsKey(attribute) && (weight = weights.get(attribute)) != null) {
                    if(weight < 0.0 || weight > 1.0) {
                        throw new IllegalArgumentException(
                                "Weights must be [0.0, 1.0]! Attribute '" + attribute + "' was " + weight);
                    }
                    resultingWeights.put(attribute, weight);
                }
            }
        }

        return resultingWeights;
    }
    
    public boolean isSimpleKNN() {
        return simpleKNN;
    }

    /**
     * Calculates the similarity between two cases.
     * Return value is normalized on the weights used, and is therefore in the range [0, 1].
     */
    @Override
    public double calculateSimilarity(Caze caze, Caze query) {
        double similarity = 0;
        double weightSum = 0;
        for(String attribute : attributes) {
            if(attribute.startsWith(Attributes.WIND_DIRECTION)) {
                Double cazeWindSpeed = null;
                Double queryWindSpeed = null;
                for(String a : attributes) {
                    if(a.startsWith(Attributes.WIND_SPEED)) {
                        cazeWindSpeed = caze.getAttributeValue(a);
                        queryWindSpeed = query.getAttributeValue(a);
                        break;
                    }
                }

                if(!ROAD_DIRECTIONS.containsKey(caze.getLocation()) || !ROAD_DIRECTIONS.containsKey(query.getLocation()) || cazeWindSpeed == null || queryWindSpeed == null || cazeWindSpeed < 2.0 || queryWindSpeed < 2.0) {
                    continue;
                }
            }

            double weight = weights.get(attribute);
            weightSum += weight;
            Double cazeValue = caze.getAttributeValue(attribute);

            Decision cazeDecision;
            if(simpleKNN) {
                cazeDecision = Decision.CONVOY;
            } else {
                cazeDecision = caze.getDecision();
            }

            Double queryValue = query.getAttributeValue(attribute);
            if(attribute.startsWith(Attributes.WIND_SPEED)) {
                similarity += weight * windSpeedSimilarity(cazeValue, cazeDecision, queryValue);
            } else if(attribute.startsWith(Attributes.WIND_DIRECTION)) {
                similarity += weight * windDirectionSimilarity(cazeValue, caze.getLocation(), cazeDecision, queryValue, query.getLocation());
            } else if(attribute.startsWith(Attributes.FRICTION)) {
                similarity += weight * frictionSimilarity(cazeValue, cazeDecision, queryValue);
            } else if(attribute.startsWith(Attributes.VISIBILITY)) {
                similarity += weight * visibilitySimilarity(cazeValue, cazeDecision, queryValue);
            } else if(attribute.startsWith(Attributes.TEMPERATURE)) {
                similarity += weight * temperatureSimilarity(cazeValue, queryValue);
            } else if(attribute.startsWith(Attributes.SPEED_LIMIT)) {
                similarity += weight * speedLimitSimilarity(cazeValue, cazeDecision, queryValue);
            } else {
                throw new RuntimeException("Unknown attribute: " + attribute);
            }

            if(Double.isNaN(similarity)) {
                System.err.printf("Similarity is NaN, likely caused by %s. Query value=%.3f, case value=%.3f\n", attribute, queryValue, cazeValue);
            }
        }

        if(Double.isNaN(similarity / weightSum)) {
            System.err.println("similarity / weightSum is NaN");
        } else if(similarity / weightSum > 1.0 || similarity / weightSum < 0.0) {
            System.err.println("similarity / weightSum is not within [0,1]: " + similarity / weightSum);
        }

        return similarity / weightSum;
    }

    public double windSpeedSimilarity(Double cazeValue, Decision cazeDecision, Double queryValue) {
        if(cazeValue == null || queryValue == null) {
            return 0.0;
        } else if(Math.abs(cazeValue - queryValue) < 1e-8) {
            return 1.0;
        }
        double fc = 1.0 / (1.0 + Math.exp(-(cazeValue - 12.5) / 5.0));
        double fq = 1.0 / (1.0 + Math.exp(-(queryValue - 12.5) / 5.0));
        double distance;
        switch(cazeDecision) {
            case OPEN:
                distance = 2.0 * (fq - fc);
                break;
            case CLOSED:
                distance = 2.0 * (fc - fq);
                break;
            default:
                distance = 2.0 * Math.abs(fq - fc);
                break;
        }

        return getSimilarityFromDistance(distance);
    }

    public double windDirectionSimilarity(Double cazeValue, String cazeLocation, Decision cazeDecision,
                                                 Double queryValue, String queryLocation) {
        if(cazeValue == null || queryValue == null) {
            return 0.0;
        } else if(Math.abs(cazeValue - queryValue) < 1e-8) {
            return 1.0;
        }
        double fc = roadRelativeAngle(cazeValue, ROAD_DIRECTIONS.get(cazeLocation));
        double fq = roadRelativeAngle(queryValue, ROAD_DIRECTIONS.get(queryLocation));
        double distance;
        switch(cazeDecision) {
            case OPEN:
                distance = (fq - fc) / 90.0;
                break;
            case CLOSED:
                distance = (fc - fq) / 90.0;
                break;
            default:
                distance = Math.abs(fq - fc) / 90.0;
                break;
        }

        return getSimilarityFromDistance(distance);
    }

    public double frictionSimilarity(Double cazeValue, Decision cazeDecision, Double queryValue) {
        if(cazeValue == null || queryValue == null) {
            return 0.0;
        } else if(Math.abs(cazeValue - queryValue) < 1e-8) {
            return 1.0;
        }
        double fc = Math.log(cazeValue + 1e-8);
        double fq = Math.log(queryValue + 1e-8);
        double distance;
        switch(cazeDecision) {
            case OPEN:
                distance = (fc - fq) / Math.log(4.0);
                break;
            case CLOSED:
                distance = (fq - fc) / Math.log(4.0);
                break;
            default:
                distance = Math.abs(fq - fc) / Math.log(4.0);
                break;
        }
        return getSimilarityFromDistance(distance);
    }

    public double visibilitySimilarity(Double cazeValue, Decision cazeDecision, Double queryValue) {
        if(cazeValue == null || queryValue == null) {
            return 0.0;
        } else if(Math.abs(cazeValue - queryValue) < 1e-8) {
            return 1.0;
        }
        double fc = Math.log(cazeValue + 1e-8);
        double fq = Math.log(queryValue + 1e-8);
        double distance;
        switch(cazeDecision) {
            case OPEN:
                distance = (fc - fq) / Math.log(4.0);
                break;
            case CLOSED:
                distance = (fq - fc) / Math.log(4.0);
                break;
            default:
                distance = Math.abs(fq - fc) / Math.log(4.0);
                break;
        }
        return getSimilarityFromDistance(distance);
    }

    public double temperatureSimilarity(Double cazeValue, Double queryValue) {
        if(cazeValue == null || queryValue == null) {
            return 0.0;
        } else if(Math.abs(cazeValue - queryValue) < 1e-8) {
            return 1.0;
        }
        double fc = 1.0 / (1.0 + Math.exp(-cazeValue / 2.0));
        double fq = 1.0 / (1.0 + Math.exp(-queryValue / 2.0));
        double distance = 2.0 * Math.abs(fq - fc);
        return getSimilarityFromDistance(distance);
    }

    public double speedLimitSimilarity(Double cazeValue, Decision cazeDecision, Double queryValue) {
        if(cazeValue == null || queryValue == null) {
            return 0.0;
        } else if(Math.abs(cazeValue - queryValue) < 1e-8) {
            return 1.0;
        }
        double fc = cazeValue;
        double fq = queryValue;
        double distance;
        switch(cazeDecision) {
            case OPEN:
                distance = (fc - fq) / 40.0;
                break;
            case CLOSED:
                distance = (fq - fc) / 40.0;
                break;
            default:
                distance = Math.abs(fq - fc) / 40.0;
                break;
        }
        return getSimilarityFromDistance(distance);
    }

    private double roadRelativeAngle(double direction, double roadDirection) {
        double angle = direction - roadDirection;
        angle = ((angle % 180.0) + 180.0) % 180.0;
        angle = Math.min(angle, 180.0 - angle);
        return angle;
    }

    /**
     * Caps the value to 0.0 <= value <= 1.0
     */
    public static double cap(double value) {
        return Math.max(0.0, Math.min(1.0, value));
    }

    /**
     * Sim(Q, C) = 1.0 - Dist(Q, C) per def
     * Ensures distance is capped to 0.0 <= distance <= 1.0
     */
    public static double getSimilarityFromDistance(double distance) {
        if(Double.isNaN(distance)) {
            return 0.0;
        }
        return 1.0 - cap(distance);
    }
}
