package no.kolkinn.cbr_dovre.new_cbr;

import jcolibri.casebase.LinealCaseBase;
import jcolibri.cbrcore.*;
import jcolibri.connector.PlainTextConnector;
import jcolibri.exception.ExecutionException;
import jcolibri.extensions.classification.ClassificationSolution;
import jcolibri.method.retrieve.NNretrieval.NNConfig;
import jcolibri.method.retrieve.NNretrieval.NNScoringMethod;
import jcolibri.method.retrieve.NNretrieval.similarity.global.Average;
import jcolibri.method.retrieve.RetrievalResult;
import jcolibri.method.retrieve.selection.SelectCases;
import jcolibri.method.reuse.classification.*;
import jcolibri.util.FileIO;
import no.kolkinn.cbr_dovre.cbr.Decision;
import no.kolkinn.cbr_dovre.new_cbr.similarity.Interval;
import no.kolkinn.cbr_dovre.new_cbr.similarity.Logarithmic;
import no.kolkinn.cbr_dovre.new_cbr.similarity.RoadDirection;
import no.kolkinn.cbr_dovre.new_cbr.similarity.Sigmoid;
import no.kolkinn.cbr_dovre.util.Formatting;

import java.util.Collection;

/**
 * Created by Mikael on 05.10.2016.
 */
public class CbrApplication implements jcolibri.cbraplications.StandardCBRApplication {

    private Connector connector;
    private CBRCaseBase caseBase;

    private static final double ROAD_DIRECTION_FOKSTUGU = 45.0;
    private static final double ROAD_DIRECTION_AVSJOEN = 60.0;
    private static final double ROAD_DIRECTION_HJERKINN = 0.0;

    private static final Sigmoid windSpeedSimilarity = new Sigmoid(12.5, 5.0, 2.0);
    private static final Logarithmic frictionSimilarity = new Logarithmic(4.0, 1.0);
    private static final Sigmoid temperatureSimilarity = new Sigmoid(0.0, 2.0, 2.0);
    private static final Logarithmic visibilitySimilarity = new Logarithmic(4.0, 1.0);
    private static final Interval speedLimitSimilarity = new Interval(40.0);

    private KNNClassificationConfig similarityConfig;

    private int correctOpen=0, correctConvoy=0, correctClosed=0,
                incorrectOpen=0, incorrectConvoy=0, incorrectClosed=0;

    public static void main(String[] args) throws ExecutionException {
        CbrApplication app = new CbrApplication();
        app.configure();
        CBRCaseBase casebase = app.preCycle();

        // Get queries
        Connector queryConnector = new PlainTextConnector();
        queryConnector.initFromXMLfile(FileIO.findFile("testsetconfig.xml"));
        CBRCaseBase queryBase = new LinealCaseBase();
        queryBase.init(queryConnector);
        Collection<CBRCase> queries = queryBase.getCases();

        for(CBRCase query : queries) {
            app.cycle(query);
        }

        app.printResults();
        app.visualiseCases();

        app.postCycle();
    }

    @Override
    public void configure() throws ExecutionException {
        connector = new PlainTextConnector();
        connector.initFromXMLfile(FileIO.findFile("trainingsetconfig.xml"));
        caseBase = new LinealCaseBase();
    }

    @Override
    public CBRCaseBase preCycle() throws ExecutionException {
        caseBase.init(connector);

        similarityConfig = (KNNClassificationConfig) getSimilarityConfig();

        return caseBase;
    }

    public NNConfig getSimilarityConfig() {
        // Configure kNN
        KNNClassificationConfig similarityConfig = new KNNClassificationConfig();
        similarityConfig.setDescriptionSimFunction(new Average());
        similarityConfig.setClassificationMethod(new SimilarityWeightedVotingMethod());
        similarityConfig.setK(3);

        // Fokstugu
        similarityConfig.addMapping(new Attribute("windSpeedFokstugu", CaseDescription.class), windSpeedSimilarity);
        similarityConfig.addMapping(new Attribute("windDirectionFokstugu", CaseDescription.class), new RoadDirection(90.0, ROAD_DIRECTION_FOKSTUGU));
        similarityConfig.addMapping(new Attribute("frictionFokstugu", CaseDescription.class), frictionSimilarity);
        similarityConfig.addMapping(new Attribute("temperatureFokstugu", CaseDescription.class), temperatureSimilarity);
        similarityConfig.addMapping(new Attribute("visibilityFokstugu", CaseDescription.class), visibilitySimilarity);
        similarityConfig.addMapping(new Attribute("speedLimitFokstugu", CaseDescription.class), speedLimitSimilarity);

        similarityConfig.setWeight(new Attribute("windSpeedFokstugu", CaseDescription.class), 0.26171875);
        similarityConfig.setWeight(new Attribute("windDirectionFokstugu", CaseDescription.class), 0.7578125);
        similarityConfig.setWeight(new Attribute("frictionFokstugu", CaseDescription.class), 0.875);
        similarityConfig.setWeight(new Attribute("temperatureFokstugu", CaseDescription.class), 0.7890625);
        similarityConfig.setWeight(new Attribute("visibilityFokstugu", CaseDescription.class), 0.94140625);
        similarityConfig.setWeight(new Attribute("speedLimitFokstugu", CaseDescription.class), 0.9453125);

        // Avsjøen
        similarityConfig.addMapping(new Attribute("windSpeedAvsjoen", CaseDescription.class), windSpeedSimilarity);
        similarityConfig.addMapping(new Attribute("windDirectionAvsjoen", CaseDescription.class), new RoadDirection(90.0, ROAD_DIRECTION_AVSJOEN));
        similarityConfig.addMapping(new Attribute("frictionAvsjoen", CaseDescription.class), frictionSimilarity);
        similarityConfig.addMapping(new Attribute("temperatureAvsjoen", CaseDescription.class), temperatureSimilarity);
        similarityConfig.addMapping(new Attribute("visibilityAvsjoen", CaseDescription.class), visibilitySimilarity);
        similarityConfig.addMapping(new Attribute("speedLimitAvsjoen", CaseDescription.class), speedLimitSimilarity);

        similarityConfig.setWeight(new Attribute("windSpeedAvsjoen", CaseDescription.class), 0.9765625);
        similarityConfig.setWeight(new Attribute("windDirectionAvsjoen", CaseDescription.class), 0.04296875);
        similarityConfig.setWeight(new Attribute("frictionAvsjoen", CaseDescription.class), 0.2265625);
        similarityConfig.setWeight(new Attribute("temperatureAvsjoen", CaseDescription.class), 0.92578125);
        similarityConfig.setWeight(new Attribute("visibilityAvsjoen", CaseDescription.class), 0.265625);
        similarityConfig.setWeight(new Attribute("speedLimitAvsjoen", CaseDescription.class), 0.3671875);

        // Hjerkinn
        similarityConfig.addMapping(new Attribute("windSpeedHjerkinn", CaseDescription.class), windSpeedSimilarity);
        similarityConfig.addMapping(new Attribute("windDirectionHjerkinn", CaseDescription.class), new RoadDirection(90.0, ROAD_DIRECTION_HJERKINN));
        similarityConfig.addMapping(new Attribute("temperatureHjerkinn", CaseDescription.class), temperatureSimilarity);
        similarityConfig.addMapping(new Attribute("visibilityHjerkinn", CaseDescription.class), visibilitySimilarity);

        similarityConfig.setWeight(new Attribute("windSpeedHjerkinn", CaseDescription.class), 0.89453125);
        similarityConfig.setWeight(new Attribute("windDirectionHjerkinn", CaseDescription.class), 0.25);
        similarityConfig.setWeight(new Attribute("temperatureHjerkinn", CaseDescription.class), 0.4375);
        similarityConfig.setWeight(new Attribute("visibilityHjerkinn", CaseDescription.class), 0.5234375);

        return similarityConfig;
    }

    @Override
    public void cycle(CBRQuery query) throws ExecutionException {

        // Execute NN
        Collection<RetrievalResult> results = NNScoringMethod.evaluateSimilarity(caseBase.getCases(), query, similarityConfig);
        Collection<RetrievalResult> knn = SelectCases.selectTopKRR(results, similarityConfig.getK());

        // Predict the solution
        KNNClassificationMethod classifier = similarityConfig.getClassificationMethod();
        ClassificationSolution predictedSolution =  classifier.getPredictedSolution(knn);

        if(predictedSolution == null || predictedSolution.getClassification() == null) {
            System.out.println(predictedSolution);
            System.out.println(query.getDescription());
            for(RetrievalResult neighbour : knn) {
                System.out.println(neighbour.toString());
            }
        }

        updateScores(query, predictedSolution);
    }

    private void updateScores(CBRQuery query, ClassificationSolution predictedSolution) {
        Decision queryDesicion = (Decision) ((CaseSolution) ((CBRCase) query).getSolution()).getClassification();
        Decision predictedDecision = (Decision) predictedSolution.getClassification();
        if(queryDesicion.equals(predictedDecision)) {
            switch(queryDesicion) {
                case OPEN:
                    correctOpen++;
                    break;
                case CONVOY:
                    correctConvoy++;
                    break;
                case CLOSED:
                    correctClosed++;
                    break;
                default:
                    throw new IllegalArgumentException("Somehow, the enum Decision in the query was null.\n"
                                                       + "Query: " + query.getDescription());
            }
        } else {
            switch(queryDesicion) {
                case OPEN:
                    incorrectOpen++;
                    break;
                case CONVOY:
                    incorrectConvoy++;
                    break;
                case CLOSED:
                    incorrectClosed++;
                    break;
                default:
                    throw new IllegalArgumentException("Somehow, the enum Decision in the query was null.\n"
                                                       + "Query: " + query.getDescription());
            }
        }
    }

    @Override
    public void postCycle() throws ExecutionException {
        connector.close();
    }

    public void printResults() {
        System.out.println(Formatting.formatResults(correctOpen, correctConvoy, correctClosed,
                                 incorrectOpen, incorrectConvoy, incorrectClosed));
    }

    public void visualiseCases() {
        visualiseCases(caseBase.getCases(), similarityConfig);
    }

    private void visualiseCases(Collection<CBRCase> cases, NNConfig similarityConfig) {
    }

}
