package no.kolkinn.cbr_dovre.tests;

import no.kolkinn.cbr_dovre.cbr.*;
import no.kolkinn.cbr_dovre.domain.SpeedLimitCalculation;
import no.kolkinn.cbr_dovre.methods.maintenance.ICFRedundancyRemoval;
import no.kolkinn.cbr_dovre.methods.maintenance.RENNNoiseReduction;
import no.kolkinn.cbr_dovre.methods.maintenance.RemoveNonCausalClosedOrConvoy;
import no.kolkinn.cbr_dovre.methods.selection.Grouping;
import no.kolkinn.cbr_dovre.models.Attributes;
import no.kolkinn.cbr_dovre.models.Model1;
import no.kolkinn.cbr_dovre.models.Model2;
import no.kolkinn.cbr_dovre.models.SimilarityFunctions;
import no.kolkinn.cbr_dovre.util.Formatting;

import java.io.File;
import java.io.IOException;
import java.text.ParseException;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.concurrent.*;
import java.util.logging.Level;
import java.util.logging.Logger;

public class Tests {

    private static final int NOF_QUERY_SETS = 1;

    public static void main(
            String[] args) throws IOException, ParseException, ExecutionException, InterruptedException {
        Logger logger = Logger.getLogger(Tests.class.getName());
        logger.info("Tests running...");
        Set<Caze> trainingSet = CazeFileReader.readFile(new File("trainingset.csv"), Attributes.ALL_LOCATIONS);
        Set<Caze> testSet = CazeFileReader.readFile(new File("testset.csv"), Attributes.ALL_LOCATIONS);
        int kNN = 3;


        Map<String, Double> model2Weights = new HashMap<>();
        model2Weights.put(Attributes.WIND_DIRECTION, 0.04);
        model2Weights.put(Attributes.TEMPERATURE, 0.88);
        model2Weights.put(Attributes.WIND_SPEED, 0.81);
        model2Weights.put(Attributes.FRICTION, 0.75);
        model2Weights.put(Attributes.SPEED_LIMIT, 0.6);
        model2Weights.put(Attributes.VISIBILITY, 0.81);

        SimilarityFunctions weightedModel2 = new Model2(model2Weights, false);
        CBR cbr = new CBR(weightedModel2, kNN);

//        Set<Caze> prune1 = new RemoveNonCausalClosedOrConvoy(
//                new CBR(new Model1(false), kNN)).removeNonCausalClosedOrConvoy(trainingSet, 24,
//                                                                             ChronoUnit.HOURS);

        Set<Caze> prune2 = new RemoveNonCausalClosedOrConvoy(
                cbr).removeNonCausalClosedOrConvoy(trainingSet, 24,
                                                   ChronoUnit.HOURS);

        System.out.println("Original");
        runTests(logger, trainingSet, testSet, kNN, NOF_QUERY_SETS, false, weightedModel2);

//        Thread.sleep(10000);
//        System.out.printf("\n\nPruned with Model1"
//                          + "\nRemoved %d cases.\n", trainingSet.size() - prune1.size());
//        runTests(logger, prune1, testSet, kNN, NOF_QUERY_SETS, new Model1(false), new Model2(false));

        Thread.sleep(10000);
        System.out.printf("\n\nOriginal pruned with Model 2"
                          + "\nRemoved %d cases.\n", trainingSet.size() - prune2.size());
        runTests(logger, prune2, testSet, kNN, NOF_QUERY_SETS, false, weightedModel2);

        RENNNoiseReduction noiseReducer = new RENNNoiseReduction(cbr);
        Set<Caze> noiseReduced = noiseReducer.reduceNoise(trainingSet);
        System.out.printf("\n\nNoise reduction on original"
                          + "\nRemoved %d cases.\n", trainingSet.size() - noiseReduced.size());
        runTests(logger, noiseReduced, testSet, kNN, NOF_QUERY_SETS, false, weightedModel2);

        Set<Caze> prune22 = new RemoveNonCausalClosedOrConvoy(
                cbr).removeNonCausalClosedOrConvoy(noiseReduced, 24,
                                                   ChronoUnit.HOURS);
        System.out.printf("\n\nNoise reduced pruned with Model 2"
                          + "\nRemoved %d cases.\n", noiseReduced.size() - prune22.size());
        runTests(logger, prune22, testSet, kNN, NOF_QUERY_SETS, false, weightedModel2);

        ICFRedundancyRemoval redundancyRemover = new ICFRedundancyRemoval(cbr);
        Set<Caze> redundancyRemoved = redundancyRemover.removeDuplicates(noiseReduced);
        System.out.printf("\n\nRedundancy removal on noiseReduced"
                          + "\nRemoved %d cases.\n", noiseReduced.size() - redundancyRemoved.size());
        runTests(logger, redundancyRemoved, testSet, kNN, NOF_QUERY_SETS, false, weightedModel2);

        Set<Caze> prune222 = new RemoveNonCausalClosedOrConvoy(
                cbr).removeNonCausalClosedOrConvoy(redundancyRemoved, 24,
                                                   ChronoUnit.HOURS);
        System.out.printf("\n\nRedundancy removed pruned with Model 2"
                          + "\nRemoved %d cases.\n", redundancyRemoved.size() - prune222.size());
        runTests(logger, prune222, testSet, kNN, NOF_QUERY_SETS, false, weightedModel2);
    }

    public static void runTests(Logger logger, Set<Caze> trainingSet,
                                Set<Caze> testSet, int kNN, int nofQuerySets, boolean onlyOtherLocations,
                                SimilarityFunctions... modelWithWeights) throws InterruptedException, ExecutionException {

        Thread.UncaughtExceptionHandler exceptionHandler = (th, ex) -> {
            logger.log(Level.SEVERE, "Uncaught exception in test pool for CBR performance evaluations.", ex);
        };
        ThreadPoolExecutor executor = new ThreadPoolExecutor(Runtime.getRuntime().availableProcessors() - 1,
                                                             Runtime.getRuntime().availableProcessors(), 3,
                                                             TimeUnit.SECONDS, new LinkedBlockingQueue<>(), r -> {
            Thread t = Executors.defaultThreadFactory().newThread(
                    r);
            t.setUncaughtExceptionHandler(exceptionHandler);
            t.setDaemon(true);
            return t;
        });

        List<Future<Results>> resultsFutures = new ArrayList<>(nofQuerySets);
        Set<Set<Caze>> timeGroupedTestCases = Grouping.groupCasesByTime(testSet);


        if(onlyOtherLocations) {
            // Create a map of sets containing cases from all locations except the map key
            List<Set<Caze>> locationGroupedTrainingCases = new ArrayList<>(
                    Grouping.groupConnectedCases(trainingSet, Long.MAX_VALUE,
                                                 ChronoUnit.DAYS));
            Map<String, Set<Caze>> trainingCasesWithoutLocation = new HashMap<>(locationGroupedTrainingCases.size());
            for(String location : Attributes.ALL_LOCATIONS) {
                trainingCasesWithoutLocation.put(location, new HashSet<>());
                for(Set<Caze> group : locationGroupedTrainingCases) {
                    if(!Objects.equals(location, group.iterator().next().getLocation())) {
                        trainingCasesWithoutLocation.get(location).addAll(group);
                    }
                }
            }

            resultsFutures.add(
                    executor.submit(new CbrTester(logger, trainingCasesWithoutLocation, timeGroupedTestCases,
                                                  modelWithWeights, kNN)));
        } else {
            resultsFutures.add(
                    executor.submit(new CbrTester(logger, trainingSet, timeGroupedTestCases, modelWithWeights, kNN)));
        }


        Results resultsAverage = new Results();
        for(int i = 0; i < resultsFutures.size(); i++) {
            Future<Results> future = resultsFutures.get(i);
            Results results = future.get();
            resultsAverage.cbr1Performance += results.cbr1Performance;
            resultsAverage.cbr2Performance += results.cbr2Performance;
            resultsAverage.cbr1MontecarloPerformance += results.cbr1MontecarloPerformance;
            resultsAverage.cbr2MontecarloPerformance += results.cbr2MontecarloPerformance;
            resultsAverage.app1Performance += results.app1Performance;
            resultsAverage.app2Performance += results.app2Performance;

            logger.info((i + 1) + " of " + nofQuerySets + " tests finished");
        }

        resultsAverage.cbr1Performance /= nofQuerySets;
        resultsAverage.cbr2Performance /= nofQuerySets;
        resultsAverage.cbr1MontecarloPerformance /= nofQuerySets;
        resultsAverage.cbr2MontecarloPerformance /= nofQuerySets;
        resultsAverage.app1Performance /= nofQuerySets;
        resultsAverage.app2Performance /= nofQuerySets;
        logger.warning("Results:"
                       + "\nCBR-model 1: " + resultsAverage.cbr1Performance
                       + "\nCBR-model 2: " + resultsAverage.cbr2Performance
                       + "\nCBR-model 1 with sensor uncertainty handling: " + resultsAverage.cbr1MontecarloPerformance
                       + "\nCBR-model 2 with sensor uncertainty handling: " + resultsAverage.cbr2MontecarloPerformance
                       + "\nApplication formula: " + resultsAverage.app1Performance
                       + "\nApplication formula ignoring missing attribute queries: " + resultsAverage.app2Performance);
    }

    public static double testCBR(Logger logger, CBR cbr, Set<Caze> cazeBase,
                                 Map<String, Set<Caze>> cazeBasesWithoutLocation, Set<Set<Caze>> timeGroupedQueries) {
        cbr.setCazeBase(cazeBase);
        int correctOpen = 0;
        int totalOpen = 0;
        int correctConvoy = 0;
        int totalConvoy = 0;
        int correctClosed = 0;
        int totalClosed = 0;

        DecisionCount decisionInQuery = new DecisionCount();
        DecisionCount decisionCount = new DecisionCount();
        for(Set<Caze> timeGroup : timeGroupedQueries) {

            // Use DecisionCount for the decision in the query, just in case
            // the decisions in it's separate cases have also been adjusted
            decisionInQuery.clear();
            decisionCount.clear();
            for(Caze query : timeGroup) {
                if(cazeBasesWithoutLocation != null) {
                    // Use a case base without cases from the location in the query
                    cbr.setCazeBase(cazeBasesWithoutLocation.get(query.getLocation()));
                    if(cbr.getCazeBase() == null) {
                        logger.severe("cazeBasesWithoutLocation did not contain " + query.getLocation() + ".\n"
                                      + cazeBasesWithoutLocation.toString());
                    }
                }

                decisionInQuery.increment(query.getDecision());
                decisionCount.increment(cbr.makeDecision(query));
            }

            switch(decisionInQuery.getDecision()) {
                case OPEN:
                    if(decisionCount.getDecision() == Decision.OPEN) {
                        correctOpen++;
                    }
                    totalOpen++;
                    break;
                case CONVOY:
                    if(decisionCount.getDecision() == Decision.CONVOY) {
                        correctConvoy++;
                    }
                    totalConvoy++;
                    break;
                case CLOSED:
                    if(decisionCount.getDecision() == Decision.CLOSED) {
                        correctClosed++;
                    }
                    totalClosed++;
                    break;
            }
        }

        int nofDecisionTypes = 0;
        double result = 0.0;
        if(totalOpen != 0) {
            nofDecisionTypes++;
            result += (double) correctOpen / (double) totalOpen;
        }
        if(totalConvoy != 0) {
            nofDecisionTypes++;
            result += (double) correctConvoy / (double) totalConvoy;
        }
        if(totalClosed != 0) {
            nofDecisionTypes++;
            result += (double) correctClosed / (double) totalClosed;
        }
        if(nofDecisionTypes == 0) {
            throw new RuntimeException("No test cases");
        }
        result /= nofDecisionTypes;

        if(logger != null) {
            logger.info("\n" + Formatting.formatResults(correctOpen, correctConvoy, correctClosed,
                                                        totalOpen - correctOpen, totalConvoy - correctConvoy,
                                                        totalClosed - correctClosed));
        }

        return result;
    }

    public static double testCBRWithMontecarloSimulation(Logger logger, CBR cbr, Set<Caze> cazeBase,
                                                         Set<Caze> queries) {
        double correctOpen = 0;
        int totalOpen = 0;
        double correctConvoy = 0;
        int totalConvoy = 0;
        double correctClosed = 0;
        int totalClosed = 0;
        for(Caze query : queries) {
            Map<Decision, Double> decisionProbabilities = MontecarloSimulation.getDecisionProbabilities(cbr, cazeBase,
                                                                                                        query);
            switch(query.getDecision()) {
                case OPEN:
                    correctOpen += decisionProbabilities.get(Decision.OPEN);
                    totalOpen++;
                    break;
                case CONVOY:
                    correctConvoy += decisionProbabilities.get(Decision.CONVOY);
                    totalConvoy++;
                    break;
                case CLOSED:
                    correctClosed += decisionProbabilities.get(Decision.CLOSED);
                    totalClosed++;
                    break;
            }
        }
        int nofDecisionTypes = 0;
        double result = 0.0;
        if(totalOpen != 0) {
            nofDecisionTypes++;
            result += correctOpen / (double) totalOpen;
        }
        if(totalConvoy != 0) {
            nofDecisionTypes++;
            result += correctConvoy / (double) totalConvoy;
        }
        if(totalClosed != 0) {
            nofDecisionTypes++;
            result += correctClosed / (double) totalClosed;
        }
        if(nofDecisionTypes == 0) {
            throw new RuntimeException("No test cases");
        }
        result /= nofDecisionTypes;

        if(logger != null) {
            logger.info("\n" + Formatting.formatResults(correctOpen, correctConvoy, correctClosed,
                                                        totalOpen - correctOpen, totalConvoy - correctConvoy,
                                                        totalClosed - correctClosed));
        }

        return result;
    }

    private static boolean convoyQueryExists(Set<Caze> queries) {
        for(Caze query : queries) {
            if(query.getDecision() == Decision.CONVOY)
                return true;
        }
        return false;
    }

    private static double testMobileApplicationFormula(Logger logger, Set<Set<Caze>> timeGroupedQueries,
                                                       boolean ignoreMissingAttributeQueries) {
        int correctOpen = 0;
        int totalOpen = 0;
        int correctConvoy = 0;
        int totalConvoy = 0;
        int correctClosed = 0;
        int totalClosed = 0;

        DecisionCount decisionInQuery = new DecisionCount();
        DecisionCount decisionCount = new DecisionCount();
        for(Set<Caze> timeGroup : timeGroupedQueries) {


            // Use DecisionCount for the decision in the query, just in case
            // the decisions in it's separate cases have also been adjusted
            decisionInQuery.clear();
            decisionCount.clear();
            for(Caze query : timeGroup) {
                decisionInQuery.increment(query.getDecision());

                Double windSpeed = query.getAttributeValue(Attributes.WIND_SPEED);
                Double friction = query.getAttributeValue(Attributes.FRICTION);
                Double visibility = query.getAttributeValue(Attributes.VISIBILITY);
                Decision decision = null;
                if(windSpeed != null && friction != null && visibility != null) {
                    decision = SpeedLimitCalculation.decision(windSpeed, friction, visibility);
                }

                if(decision != null) {
                    decisionCount.increment(decision);
                }
            }

            if(decisionCount.getDecision() == null && ignoreMissingAttributeQueries) {
                continue;
            }

            switch(decisionInQuery.getDecision()) {
                case OPEN:
                    if(decisionCount.getDecision() == Decision.OPEN) {
                        correctOpen++;
                    }
                    totalOpen++;
                    break;
                case CONVOY:
                    if(decisionCount.getDecision() == Decision.CONVOY) {
                        correctConvoy++;
                    }
                    totalConvoy++;
                    break;
                case CLOSED:
                    if(decisionCount.getDecision() == Decision.CLOSED) {
                        correctClosed++;
                    }
                    totalClosed++;
                    break;
            }
        }

        int nofDecisionTypes = 0;
        double result = 0.0;
        if(totalOpen != 0) {
            nofDecisionTypes++;
            result += (double) correctOpen / (double) totalOpen;
        }
        if(totalConvoy != 0) {
            nofDecisionTypes++;
            result += (double) correctConvoy / (double) totalConvoy;
        }
        if(totalClosed != 0) {
            nofDecisionTypes++;
            result += (double) correctClosed / (double) totalClosed;
        }
        if(nofDecisionTypes == 0) {
//            throw new RuntimeException("No test cases");
            return -1.0;
        }
        result /= nofDecisionTypes;

        if(logger != null) {
            logger.info("\n" + Formatting.formatResults(correctOpen, correctConvoy, correctClosed,
                                                        totalOpen - correctOpen, totalConvoy - correctConvoy,
                                                        totalClosed - correctClosed));
        }

        return result;
    }

    private static class CbrTester implements Callable<Results> {
        private final Logger logger;
        private final Map<String, Set<Caze>> cazeBasesWithoutLocation;
        private final Set<Caze> cazeBase;
        private final Set<Set<Caze>> queries;
        private final SimilarityFunctions[] modelWithWeights;
        private final int kNN;

        /**
         * Constructor for ordinary testing.
         * Here all locations are used to try and find a
         * solution to a query.
         */
        public CbrTester(Logger logger, Set<Caze> cazeBase, Set<Set<Caze>> queries,
                         SimilarityFunctions[] modelWithWeights,
                         int kNN) {
            this.logger = logger;
            this.cazeBase = cazeBase;
            this.cazeBasesWithoutLocation = null;
            this.queries = queries;
            this.modelWithWeights = modelWithWeights;
            this.kNN = kNN;
        }

        /**
         * Constructor for testing the performance of cases
         * from other locations than the location of a query.
         */
        public CbrTester(Logger logger, Map<String, Set<Caze>> cazeBasesWithoutLocation,
                         Set<Set<Caze>> queries, SimilarityFunctions[] modelWithWeights, int kNN) {

            this.logger = logger;
            this.cazeBasesWithoutLocation = cazeBasesWithoutLocation;
            this.cazeBase = null;
            this.queries = queries;
            this.modelWithWeights = modelWithWeights;
            this.kNN = kNN;
        }

        @Override
        public Results call() throws Exception {
            Results results = new Results();
            for(SimilarityFunctions model : modelWithWeights) {
                CBR cbr = new CBR(model, kNN);

                if(model instanceof Model1) {
                    results.cbr1Performance = testCBR(logger, cbr, cazeBase, cazeBasesWithoutLocation, queries);
//                    results.cbr1MontecarloPerformance = testCBRWithMontecarloSimulation(logger, cbr, cazeBase, queries);
                } else {
                    results.cbr2Performance = testCBR(logger, cbr, cazeBase, cazeBasesWithoutLocation, queries);
//                    results.cbr2MontecarloPerformance = testCBRWithMontecarloSimulation(logger, cbr, cazeBase, queries);
                }
            }
//            results.app1Performance = testMobileApplicationFormula(logger, queries, false);
//            results.app2Performance = testMobileApplicationFormula(logger, queries, true);

            return results;
        }
    }

    private static class Results {
        double cbr1Performance = 0.0;
        double cbr2Performance = 0.0;
        double cbr1MontecarloPerformance = 0.0;
        double cbr2MontecarloPerformance = 0.0;
        double app1Performance = 0.0;
        double app2Performance = 0.0;
    }

    private static class DecisionCount {
        private Map<Decision, Integer> count;

        DecisionCount() {
            count = new HashMap<>(Decision.values().length);
            clear();
        }

        void increment(Decision decision) {
            count.compute(decision, (k, v) -> v + 1);
        }

        int getCount(Decision decision) {
            return count.get(decision);
        }

        Decision getDecision() {
            if(getCount(Decision.CLOSED) > 0) {
                return Decision.CLOSED;
            } else if(getCount(Decision.CONVOY) > 0) {
                return Decision.CONVOY;
            } else if(getCount(Decision.OPEN) > 0) {
                return Decision.OPEN;
            } else {
                return null;
            }
        }

        void clear() {
            for(Decision decision : Decision.values()) {
                count.put(decision, 0);
            }
        }
    }
}
