package no.kolkinn.cbr_dovre.visual.star_coordinates;

import com.sun.glass.ui.Application;
import com.sun.javafx.charts.Legend;
import com.sun.javafx.collections.ObservableListWrapper;
import javafx.application.Platform;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.MapProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.beans.property.SimpleMapProperty;
import javafx.collections.FXCollections;
import javafx.collections.MapChangeListener;
import javafx.collections.ObservableList;
import javafx.collections.ObservableMap;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.Initializable;
import javafx.scene.Cursor;
import javafx.scene.control.*;
import javafx.scene.control.cell.CheckBoxListCell;
import javafx.scene.layout.VBox;
import javafx.stage.FileChooser;
import javafx.util.StringConverter;
import no.kolkinn.cbr_dovre.cbr.Caze;
import no.kolkinn.cbr_dovre.cbr.CazeFileReader;
import no.kolkinn.cbr_dovre.cbr.Decision;
import no.kolkinn.cbr_dovre.models.Attributes;
import no.kolkinn.cbr_dovre.visual.star_coordinates.chart.StarCoordinatesChart;

import java.io.File;
import java.net.URL;
import java.util.*;
import java.util.concurrent.*;

/**
 * Created by Mikael on 29.10.2016.
 */
public class ChartController implements Initializable {

    @FXML
    private VBox rightVBox;
    @FXML
    private ListView<StarCoordinatesChart.UnitVector> attributesListView;
    @FXML
    private Accordion selectedCasesAccordion;
    @FXML
    private StarCoordinatesChart chart;

    private Map<String, String> locationStyle = new HashMap<>();
    private Map<Decision, String> decisionStyle = new HashMap<>();
    private FileChooser fileChooser;
    private ExecutorService executor = Executors.newFixedThreadPool(2, runnable -> {
        Thread thread = Executors.defaultThreadFactory().newThread(runnable);

        // Exits if the app quits
        thread.setDaemon(true);
        return thread;
    });

    /**
     * Map of cases with {@link ChartController#getName(Caze)} used as the key.
     */
    private MapProperty<String, Caze> cases = new SimpleMapProperty<>(FXCollections.synchronizedObservableMap(FXCollections.observableHashMap()));
    private MapProperty<String, StarCoordinatesChart.UnitVector> unitVectors = new SimpleMapProperty<>(
            FXCollections.synchronizedObservableMap(FXCollections.observableHashMap()));

    {
        decisionStyle.put(Decision.OPEN, "open");
        decisionStyle.put(Decision.CONVOY, "convoy");
        decisionStyle.put(Decision.CLOSED, "closed");
    }

    @Override
    protected void finalize() throws Throwable {
        executor.shutdown();
        super.finalize();
    }

    @Override
    public void initialize(URL location, ResourceBundle resources) {
        fileChooser = new FileChooser();

        // Show vector.getName() in cells and add vector to chart if selected
        // and remove it if deselected.
        // Cell check boxes default to selected=true
        attributesListView.setCellFactory(CheckBoxListCell.forListView(
                vector -> {
                    BooleanProperty selected = new SimpleBooleanProperty(true);
                    selected.addListener((observable, oldValue, newValue) -> {
                        if(newValue) {
                            chart.getUnitVectors().put(vector.getName(), vector);
                        } else {
                            chart.getUnitVectors().remove(vector.getName());
                        }
                    });

                    return selected;
                }, new StringConverter<StarCoordinatesChart.UnitVector>() {
                    @Override
                    public String toString(StarCoordinatesChart.UnitVector vector) {
                        return vector.getName();
                    }

                    @Override
                    public StarCoordinatesChart.UnitVector fromString(String name) {
                        return unitVectors.get(name);
                    }
                }));


        // Have attributesListView bind to the values in unitVectors
        unitVectors.addListener((MapChangeListener<String, StarCoordinatesChart.UnitVector>) change -> {
            final boolean wasAdded = change.wasAdded();
            final StarCoordinatesChart.UnitVector vectorRemoved = change.getValueRemoved();
            final StarCoordinatesChart.UnitVector vectorAdded = change.getValueAdded();
            if(Platform.isFxApplicationThread()) {
                updateAttributesListView(wasAdded, vectorRemoved, vectorAdded);
            } else {
                Platform.runLater(() -> updateAttributesListView(wasAdded, vectorRemoved, vectorAdded));
            }
        });
    }

    private void updateAttributesListView(boolean wasAdded, StarCoordinatesChart.UnitVector valueRemoved,
                                          StarCoordinatesChart.UnitVector valueAdded) {
        attributesListView.getItems().removeAll(valueRemoved);
        if(wasAdded) {
            attributesListView.getItems().add(valueAdded);
        }
    }

    public void resetAxes(ActionEvent actionEvent) {
        chart.resetViewAndVectorPosition();
    }

    public void loadCasesFile(ActionEvent actionEvent) {
        fileChooser = new FileChooser();
        fileChooser.setTitle("Select cases file");
        File file = fileChooser.showOpenDialog(((Button) actionEvent.getSource()).getScene().getWindow());
        Set<Caze> cases;

        if(file == null) {
            return;
        }
        try {
            cases = CazeFileReader.readFile(file, Attributes.ALL_LOCATIONS);
            if(cases == null || cases.isEmpty()) {
                throw new Exception("File contained no cases or was not a valid CSV file.");
            }
        } catch(Exception e) {
            Alert alert = new Alert(Alert.AlertType.ERROR);
            alert.setTitle("Error loading file");
            alert.setHeaderText("An error occurred when trying to load file " + file.getName() + ".");
            alert.setContentText(e.getLocalizedMessage());
            alert.show();

            e.printStackTrace();
            return;
        }

        loadCases(cases);
    }

    public Future<Boolean> setUnitVectors(Set<Caze> caseSet) {
        unitVectors.clear();
        Callable<Boolean> callable = Executors.callable(() -> {
            for(Caze caze : caseSet) {
                for(String attribute : caze.getAttributes()) {
                    if(caze.getAttributeValue(attribute) != null) {
                        unitVectors.computeIfAbsent(attribute, StarCoordinatesChart.UnitVector::new);
                    }
                }
            }
        }, true);
        return executor.submit(callable);
    }

    private void loadCases(Set<Caze> caseSet) {
        chart.getUnitVectors().clear();
        chart.getData().clear();
        selectedCasesAccordion.getPanes().clear();
        cases.clear();

        Future<Boolean> unitVectorFuture = setUnitVectors(caseSet);

        Future<ObservableList<StarCoordinatesChart.Data>> dataItemFuture = executor.submit(() -> {
            ObservableList<StarCoordinatesChart.Data> dataItems = FXCollections.observableArrayList();
            caseSet.forEach(caze -> {
                String key = getName(caze);
                cases.put(key, caze);

                // 0..8 styles in main.css
                locationStyle.computeIfAbsent(caze.getLocation(), s -> "location" + locationStyle.size() % 9);

                ObservableMap<String, Double> attributeValues = FXCollections.observableHashMap();
                for(String attribute : caze.getAttributes()) {
                    attributeValues.put(attribute, caze.getAttributeValue(attribute));
                }

                StarCoordinatesChart.Data data = new StarCoordinatesChart.Data(
                        key,
                        attributeValues,
                        locationStyle.get(caze.getLocation()),
                        decisionStyle.get(caze.getDecision())
                );

                setSelectedListener(data);
                dataItems.add(data);
            });

            return dataItems;
        });


        try {
            unitVectorFuture.get();
            chart.setData(dataItemFuture.get());
        } catch(InterruptedException | ExecutionException e) {
            e.printStackTrace();
        }

        // Use a copy of unitVectors content, so that
        // chart cannot remove elements in our list
        chart.getUnitVectors().putAll(unitVectors);
        chart.defaultUnitVectorPositions();
        setLegend();
    }

    private void setLegend() {
        Legend legend = new Legend();
        for(Map.Entry<String, String> entry : locationStyle.entrySet()) {
            Legend.LegendItem item = new Legend.LegendItem(entry.getKey());
            item.getSymbol().getStyleClass().setAll("chart-data-point", "chart-symbol", "no-color", entry.getValue());
            legend.getItems().add(item);
        }
        for(Map.Entry<Decision, String> entry : decisionStyle.entrySet()) {
            Legend.LegendItem item = new Legend.LegendItem(entry.getKey().toString());
            item.getSymbol().getStyleClass().setAll("chart-data-point", "chart-symbol", entry.getValue());
            legend.getItems().add(item);
        }
        chart.setLegend(legend);
    }

    private ObservableList<StarCoordinatesChart.Data> getDataList() {
        ObservableList<StarCoordinatesChart.Data> data;
        data = new ObservableListWrapper<>(new ArrayList<>());
        cases.forEach((key, caze) -> {
            ObservableMap<String, Double> attributeValues = FXCollections.observableHashMap();
            for(String attribute : caze.getAttributes()) {
                attributeValues.put(attribute, caze.getAttributeValue(attribute));
            }
            data.add(new StarCoordinatesChart.Data(
                    key,
                    attributeValues,
                    locationStyle.get(caze.getLocation()),
                    decisionStyle.get(caze.getDecision())
            ));
        });
        return data;
    }

    private void setSelectedListener(StarCoordinatesChart.Data item) {
        item.selectedProperty().addListener((observable, oldValue, newValue) -> {
            if(newValue) {
                addDetailsPane(cases.get(item.getName()));
            } else {
                removeDetailsPane(cases.get(item.getName()));
            }
        });
    }

    private void addDetailsPane(Caze caze) {
        ListView<String> detailsList = new ListView<>(FXCollections.observableArrayList());
        detailsList.getItems().add("Time: " + caze.getTimestamp());
        detailsList.getItems().add("Location: " + caze.getLocation());
        detailsList.getItems().add("Decision: " + caze.getDecision().toString());

        for(String attribute : caze.getAttributes()) {
            detailsList.getItems().add(attribute + ": " + caze.getAttributeValue(attribute));
        }

        TitledPane detailsPane = new TitledPane(getName(caze), detailsList);
        selectedCasesAccordion.getPanes().add(detailsPane);
        selectedCasesAccordion.setExpandedPane(detailsPane);
    }

    private void removeDetailsPane(Caze caze) {
        selectedCasesAccordion.getPanes().removeIf(titledPane -> getName(caze).equals(titledPane.getText()));
    }

    private String getName(Caze caze) {
        return caze.getTimestamp().toString() + " " + caze.getLocation();
    }
}
