package no.kolkinn.cbr_dovre.visual.star_coordinates.chart;

import com.sun.javafx.charts.Legend;
import com.sun.javafx.collections.NonIterableChange;
import javafx.application.Platform;
import javafx.beans.property.*;
import javafx.collections.*;
import javafx.geometry.Bounds;
import javafx.geometry.Point2D;
import javafx.scene.Cursor;
import javafx.scene.Node;
import javafx.scene.chart.Chart;
import javafx.scene.input.MouseButton;
import javafx.scene.layout.StackPane;
import javafx.scene.shape.LineTo;
import javafx.scene.shape.MoveTo;
import javafx.scene.shape.Path;
import javafx.scene.text.Text;

import java.net.URL;
import java.util.*;
import java.util.concurrent.ThreadLocalRandom;

/**
 * Created by Mikael on 30.10.2016.
 */
public class StarCoordinatesChart extends Chart {

    private static final double DEFAULT_UNIT_VECTOR_LENGTH = 10.0;
    private static final double DEFAULT_SCALE_VALUE = 10.0;
    private static final double MIN_SCALE_VALUE = 0.05;

    private Legend legend = new Legend();

    private Map<String, Double> minValues = new HashMap<>();
    private Map<String, Double> maxValues = new HashMap<>();

    private MapChangeListener<String, UnitVector> unitVectorChangeListener = change -> {
        if(change.wasRemoved()) {
            unitVectorRemoved(change.getValueRemoved());
        } else if(change.wasAdded()) {
            unitVectorAdded(change.getValueAdded());
        }

        // re-layout everything
        requestChartLayout();
    };

    /**
     * Used for panning around
     */
    private DoubleProperty centerX = new DoublePropertyBase() {

        @Override
        protected void invalidated() {
            requestChartLayout();
        }

        @Override
        public Object getBean() {
            return StarCoordinatesChart.this;
        }

        @Override
        public String getName() {
            return "centerX";
        }
    };

    /**
     * Used for panning around
     */
    private DoubleProperty centerY = new DoublePropertyBase() {

        @Override
        protected void invalidated() {
            requestChartLayout();
        }

        @Override
        public Object getBean() {
            return StarCoordinatesChart.this;
        }

        @Override
        public String getName() {
            return "centerY";
        }
    };

    /**
     * Used for zooming
     */
    private DoubleProperty scale = new DoublePropertyBase(DEFAULT_SCALE_VALUE) {

        @Override
        protected void invalidated() {
            requestChartLayout();
        }

        @Override
        public Object getBean() {
            return StarCoordinatesChart.this;
        }

        @Override
        public String getName() {
            return "scale";
        }

        @Override
        public void setValue(Number v) {
            if(v != null && v.doubleValue() < MIN_SCALE_VALUE) {
                super.setValue(MIN_SCALE_VALUE);
            } else {
                super.setValue(v);
            }
        }
    };

    private MapProperty<String, UnitVector> unitVectors = new MapPropertyBase<String, UnitVector>() {

        private ObservableMap<String, UnitVector> old;

        @Override
        public Object getBean() {
            return StarCoordinatesChart.this;
        }

        @Override
        protected void invalidated() {
            final ObservableMap<String, UnitVector> current = getValue();

            if(old == current) {
                // Same map, but content has changed.
                // Let the unitVectorChangeListener do the work
                return;
            }

            // add remove listeners
            if(old != null) {
                old.removeListener(unitVectorChangeListener);
            }
            if(current != null) {
                current.addListener(unitVectorChangeListener);
            }

            // fire data change event if series are added or removed
            if(old != null) {
                for(Entry<String, UnitVector> entry : old.entrySet()) {
                    unitVectorChangeListener.onChanged(new MapChangeListener.Change<String, UnitVector>(old) {
                        @Override
                        public boolean wasAdded() {
                            return false;
                        }

                        @Override
                        public boolean wasRemoved() {
                            return true;
                        }

                        @Override
                        public String getKey() {
                            return entry.getKey();
                        }

                        @Override
                        public UnitVector getValueAdded() {
                            return null;
                        }

                        @Override
                        public UnitVector getValueRemoved() {
                            return entry.getValue();
                        }
                    });
                }
            }

            if(current != null) {
                for(Entry<String, UnitVector> entry : current.entrySet()) {
                    unitVectorChangeListener.onChanged(new MapChangeListener.Change<String, UnitVector>(current) {
                        @Override
                        public boolean wasAdded() {
                            return true;
                        }

                        @Override
                        public boolean wasRemoved() {
                            return false;
                        }

                        @Override
                        public String getKey() {
                            return entry.getKey();
                        }

                        @Override
                        public UnitVector getValueAdded() {
                            return entry.getValue();
                        }

                        @Override
                        public UnitVector getValueRemoved() {
                            return null;
                        }
                    });
                }
            }

            old = current;
        }

        @Override
        public String getName() {
            return "unitVectors";
        }
    };

    private ListProperty<Data> data = new ListPropertyBase<Data>() {

        private ObservableList<Data> old;

        @Override
        protected void invalidated() {
            final ObservableList<Data> current = getValue();

            if(old == current) {
                // Same map, but content has changed.
                // Let the unitVectorChangeListener do the work
                return;
            }

            // add remove listeners
            if(old != null) {
                old.removeListener(dataChangeListener);
            }
            if(current != null) {
                current.addListener(dataChangeListener);
            }

            // fire data change event if series are added or removed
            if(old != null || current != null) {
                final List<Data> removed = (old != null) ? old : Collections.emptyList();
                final int toIndex = (current != null) ? current.size() : 0;
                // let data listener know all old data have been removed and new data that has been added
                if(toIndex > 0 || !removed.isEmpty()) {
                    dataChangeListener.onChanged(new NonIterableChange<Data>(0, toIndex, current) {
                        @Override
                        public List<Data> getRemoved() {
                            return removed;
                        }

                        @Override
                        protected int[] getPermutation() {
                            return new int[0];
                        }

                        @Override
                        public boolean wasPermutated() {
                            return false;
                        }


                    });
                }
            } else if(old != null && old.size() > 0) {
                // FIXME: This likely won't ever get called, if so remove
                System.err.println("Part that shouldn't have been called was called...");

                // let series listener know all old series have been removed
                dataChangeListener.onChanged(new NonIterableChange<Data>(0, 0, current) {
                    @Override
                    public List<Data> getRemoved() {
                        return old;
                    }

                    @Override
                    public boolean wasPermutated() {
                        return false;
                    }

                    @Override
                    protected int[] getPermutation() {
                        return new int[0];
                    }
                });
            }

            old = current;
        }

        @Override
        public Object getBean() {
            return StarCoordinatesChart.this;
        }

        @Override
        public String getName() {
            return "data";
        }
    };

    private ListChangeListener<Data> dataChangeListener = change -> {
        while(change.next()) {
            if(change.wasPermutated()) {
                // The list was probably just sorted. Return
                return;
            }

            change.getRemoved().forEach(this::dataItemRemoved);

            for(int i = change.getFrom(); i < change.getTo(); i++) {
                Data item = getData().get(i);
                dataItemAdded(item);
            }
        }

        // re-layout everything
        requestChartLayout();
    };

    /**
     * Construct a new empty StarCoordinatesChart.
     */
    public StarCoordinatesChart() {
        this(FXCollections.observableHashMap(), FXCollections.observableArrayList());
    }

    public StarCoordinatesChart(ObservableMap<String, UnitVector> unitVectors, ObservableList<Data> data) {
        // Animation is not implemented
        setAnimated(false);

        // Add default style
        try {
            URL stylesheet = getClass().getResource("chart.css");
            getStylesheets().add(stylesheet.toString());
        } catch(Exception e) {
            // Could not load the default style, continue without it
            e.printStackTrace();
        }

        setLegend(legend);
        setUnitVectors(unitVectors);
        defaultUnitVectorPositions();
        setData(data);

        makeZoomAndPannable();
    }

    private void makeZoomAndPannable() {
        Platform.runLater(() -> {
            final Delta dragDelta = new Delta();
            getScene().setOnScroll(se -> {
                getScene().setCursor(Cursor.V_RESIZE);
                scale.setValue(scale.get() + se.getDeltaY() * 0.01);
                getScene().setCursor(Cursor.DEFAULT);
            });

            getScene().setOnMousePressed(me -> {
                if(me.isSecondaryButtonDown()) {
                    getScene().setCursor(Cursor.MOVE);
                    dragDelta.x = me.getSceneX() - centerX.getValue();
                    dragDelta.y = me.getSceneY() - centerY.getValue();
                }
            });
            getScene().setOnMouseDragged(me -> {
                if(me.isSecondaryButtonDown()) {
                    centerX.setValue(me.getSceneX() - dragDelta.x);
                    centerY.setValue(me.getSceneY() - dragDelta.y);
                }
            });
            getScene().setOnMouseReleased(me -> {
                if(!me.isSecondaryButtonDown()) {
                    getScene().setCursor(Cursor.DEFAULT);
                }
            });
        });
    }

    public void setLegend(Legend legend) {
        super.setLegend(legend);
    }

    public void resetViewAndVectorPosition() {
        centerX.setValue(0.0);
        centerY.setValue(0.0);
        scale.setValue(DEFAULT_SCALE_VALUE);
        defaultUnitVectorPositions();
    }

    /**
     * Distributes all unit vectors evenly and resets their length
     */
    public void defaultUnitVectorPositions() {
        double deltaAngle = Math.PI * 2.0 / getUnitVectors().size();
        double angle = 0.0;

        for(UnitVector vector : getUnitVectors().values()) {
            vector.setXValue(calcX(angle, DEFAULT_UNIT_VECTOR_LENGTH));
            vector.setYValue(calcY(angle, DEFAULT_UNIT_VECTOR_LENGTH));
            angle += deltaAngle;
        }
    }

    public ObservableMap<String, UnitVector> getUnitVectors() {
        return unitVectors.getValue();
    }

    /**
     * Note: angle in radians
     */
    private static double calcX(double angle, double radius) {
        return radius * Math.cos(angle);
    }

    /**
     * Note: angle in radians
     */
    private static double calcY(double angle, double radius) {
        return radius * Math.sin(angle);
    }

    public void setUnitVectors(ObservableMap<String, UnitVector> value) {
        unitVectors.setValue(value);
    }

    public MapProperty<String, UnitVector> unitVectorsProperty() {
        return unitVectors;
    }

    public ListProperty<Data> dataProperty() {
        return data;
    }

    /**
     * @inheritDoc
     */
    @Override
    protected void layoutChartChildren(double top, double left, double width, double height) {
        double pannedCenterX = width / 2 + left + centerX.getValue();
        double pannedCenterY = height / 2 + top + centerY.getValue();

        for(Data item : getData()) {
            Node node = item.getNode();
            if(node != null) {
                final double x = scaleValue(item.getXValue()) + pannedCenterX;
                final double y = scaleValue(item.getYValue()) + pannedCenterY;
                moveCentered(node, x, y);

                // Labels
                if(item.textNode.isVisible()) {
                    // Remove any scale on the text node
                    item.textNode.getTransforms().clear();

                    // Move label so that the bottom of the label
                    // is above node.
                    final double w = item.textNode.prefWidth(-1);
                    final double h = item.textNode.prefHeight(-1);
                    item.textNode.resizeRelocate(node.getLayoutX(), node.getLayoutY() - h, w, h);
                }
            }
        }

        for(UnitVector vector : getUnitVectors().values()) {
            Node node = vector.getNode();

            if(node != null) {
                // If vector is being moved by hand, don't interrupt and reset it's position
                final double nodeX = scaleValue(vector.getXValue()) + pannedCenterX;
                final double nodeY = scaleValue(vector.getYValue()) + pannedCenterY;
                moveCentered(node, nodeX, nodeY);

                // Rotate shape relative to angle from center
                node.setRotate(calcAngle(node, pannedCenterX, pannedCenterY));

                // Vector line
                vector.vectorLine.getElements().clear();
                vector.vectorLine.getElements().add(new MoveTo(pannedCenterX, pannedCenterY));
                vector.vectorLine.getElements().add(new LineTo(nodeX, nodeY));

                // Labels
                if(vector.textNode.isVisible()) {
                    Bounds nodeBounds = node.getLayoutBounds();
                    Bounds textNodeBounds = vector.textNode.getLayoutBounds();
                    final double nodeCenterX = node.getLayoutX() + nodeBounds.getWidth() / 2;
                    final double nodeCenterY = node.getLayoutY() + nodeBounds.getHeight() / 2;
                    Point2D originToNode = new Point2D(nodeCenterX, nodeCenterY).subtract(pannedCenterX, pannedCenterY);

                    // Ensure all of the text is on the outside of the UnitVector
                    double x, y;
                    if(originToNode.getX() < 0) {
                        x = nodeCenterX - (nodeBounds.getWidth() / 2) - textNodeBounds.getWidth();
                    } else {
                        x = nodeCenterX + (nodeBounds.getWidth() / 2);
                    }
                    if(originToNode.getY() < 0) {
                        y = nodeCenterY - (nodeBounds.getHeight() / 2) - textNodeBounds.getHeight();
                    } else {
                        y = nodeCenterY + (nodeBounds.getHeight() / 2);
                    }
                    vector.textNode.resizeRelocate(x, y, textNodeBounds.getWidth(), textNodeBounds.getHeight());
                }
            }
        }
    }

    private double scaleValue(double value) {
        return value * scale.getValue();
    }

    private double deScaleValue(double value) {
        return value / scale.getValue();
    }

    public ObservableList<Data> getData() {
        return data.getValue();
    }

    /**
     * Moved the node with it's center to the given position.
     */
    private void moveCentered(Node node, double x, double y) {
        final double w = node.prefWidth(-1);
        final double h = node.prefHeight(-1);
        node.resizeRelocate(x - (w / 2), y - (h / 2), w, h);
    }

    /**
     * Returns angle from origin (x, y) to node.
     * north ->   0,
     * east  ->  90,
     * south -> 180,
     * west  -> 270
     */
    private double calcAngle(Node shape, double x, double y) {
        final Point2D origin = new Point2D(x, y);
        final Point2D point = new Point2D(shape.getLayoutX(), shape.getLayoutY());
        final Point2D originToPoint = point.subtract(origin);
        double n = 180.0 - (Math.atan2(originToPoint.getX(), originToPoint.getY()) * 180.0 / Math.PI);
        return n % 360;
    }

    private Point2D getCenter(Node node) {
        Bounds bounds = node.getBoundsInParent();
        return new Point2D(
                bounds.getMinX() + (bounds.getMaxX() - bounds.getMinX()) / 2,
                bounds.getMinY() + (bounds.getMaxY() - bounds.getMinY()) / 2
        );
    }

    public void setData(ObservableList<Data> value) {
        data.setValue(value);
    }

    private void dataNameChanged(Data item) {
        item.textNode.setText(item.getName());
        requestChartLayout();
    }

    private void dataValueChanged(Data item) {
        updateEdgeValues();

        item.setXValue(calcX(item));
        item.setYValue(calcY(item));

        // TODO
        if(shouldAnimate()) {

        } else {
            requestChartLayout();
        }
    }

    private void updateEdgeValues() {
        minValues.clear();
        maxValues.clear();
        for(String attribute : getUnitVectors().keySet()) {
            minValues.put(attribute, findMinValue(attribute));
            maxValues.put(attribute, findMaxValue(attribute));
        }
    }

    private double calcX(Data item) {
        double x = 0.0;
        if(!getUnitVectors().isEmpty()) {
            for(String key : getUnitVectors().keySet()) {

                // Only bother about attributes the item has
                if(item.getAttributeValues().containsKey(key) && item.getAttributeValues().get(key) != null) {
                    x += getUnitVectors().get(key).getXValue() / (maxValues.get(key) - minValues.get(key))
                         * (item.getAttributeValues().get(key) - minValues.get(key));
                }
            }
        }
        return x;
    }

    private double calcY(Data item) {
        double y = 0.0;
        if(!getUnitVectors().isEmpty()) {
            for(String key : getUnitVectors().keySet()) {

                // Only bother about attributes the item has
                if(item.getAttributeValues().containsKey(key) && item.getAttributeValues().get(key) != null) {
                    y += getUnitVectors().get(key).getYValue() / (maxValues.get(key) - minValues.get(
                            key)) * (item.getAttributeValues().get(key) - minValues.get(key));
                }
            }
        }
        return y;
    }

    private double findMinValue(String attribute) {
        OptionalDouble optionalMinimum = getData().stream().mapToDouble(
                item -> doubleOrDefault(item.getAttributeValues().get(attribute), Double.POSITIVE_INFINITY)).min();

        if(optionalMinimum.isPresent() && !Double.isNaN(optionalMinimum.getAsDouble())) {
            return optionalMinimum.getAsDouble();
        } else {
            return 0.0; // TODO: Check if another value or behaviour is more appropriate
        }
    }

    private double findMaxValue(String attribute) {
        OptionalDouble optionalMaximum = getData().stream().mapToDouble(
                item -> doubleOrDefault(item.getAttributeValues().get(attribute), 0.0)).max();

        if(optionalMaximum.isPresent() && !Double.isNaN(optionalMaximum.getAsDouble())) {
            return optionalMaximum.getAsDouble();
        } else {
            return 1.0; // TODO: Check if another value or behaviour is more appropriate
        }
    }

    /**
     * Returns the double value of {@code value} if it can be converted to one
     * and is not NaN. Otherwise, the {@code defaultValue} is returned.
     */
    private static double doubleOrDefault(Object value, double defaultValue) {
        if(value != null) {
            if(value instanceof Double) {
                Double v = (Double) value;
                if(!v.isNaN()) {
                    return v;
                }
            } else if(value instanceof Number) {
                double v = ((Number) value).doubleValue();
                if(!Double.isNaN(v)) {
                    return v;
                }
            }
        }
        return defaultValue;
    }

    private void unitVectorChanged(UnitVector vector) {
        updateAllDataValues();
    }

    private void updateAllDataValues() {

        // Update edge values just in case there is
        // a race condition between the data change
        // listener and the unit vector change listener.
        updateEdgeValues();

        for(Data item : getData()) {
            item.setXValue(calcX(item));
            item.setYValue(calcY(item));
        }

        if(shouldAnimate()) {
            // TODO
        } else {

            requestChartLayout();
        }
    }

    private void dataItemRemoved(Data item) {
        Node shape = item.getNode();
        if(shouldAnimate()) {
            // TODO
        } else {
            getChartChildren().remove(item.textNode);
            getChartChildren().remove(shape);

            // Remove chart references from old data
            item.setChart(null);
            item.setNode(null);
        }

        if(hasEdgeValue(item)) {
            updateEdgeValues();
        }
    }

    /**
     * True if at least one of {@code item}'s attribute values is the
     * same as or "more edgy" than the existing edge value for that attribute.
     */
    private boolean hasEdgeValue(Data item) {
        for(Map.Entry<String, Double> entry : item.getAttributeValues().entrySet()) {
            if(entry.getValue() != null && getUnitVectors().containsKey(entry.getKey())) {
                if(!minValues.containsKey(entry.getKey()) || !maxValues.containsKey(entry.getKey())) {
                    updateEdgeValues();
                    return false;
                }

                if(entry.getValue() <= minValues.get(entry.getKey())) {
                    return true;
                } else if(entry.getValue() >= maxValues.get(entry.getKey())) {
                    return true;
                }
            }
        }

        return false;
    }

    private void unitVectorRemoved(UnitVector vector) {
        updateAllDataValues();
        Node shape = vector.getNode();

        if(shouldAnimate()) {
            // TODO
        } else {
            getChartChildren().remove(vector.vectorLine);
            getChartChildren().remove(vector.textNode);
            getChartChildren().remove(shape);

            // Remove chart references from old data
            vector.setChart(null);
            vector.setNode(null);
        }
    }

    private void dataItemAdded(Data item) {
        if(hasEdgeValue(item)) {
            updateEdgeValues();
        }

        item.setChart(StarCoordinatesChart.this);
        item.setXValue(calcX(item));
        item.setYValue(calcY(item));

        Node shape = createNewDataItemSymbol(item);
        Text text = createDataItemLabel(item);

        // Add toggle for "selected mode" where the label is visible
        shape.setOnMouseClicked(me -> {
            if(me.getButton().equals(MouseButton.PRIMARY)) {
                item.setSelected(!item.getSelected());
            }
        });

        getChartChildren().add(shape);
        getChartChildren().add(text);
        orderChartChildren();   // Make sure text and unit vectors are on top

        if(shouldAnimate()) {
            // TODO gradually increase opacity or something like that
        }
    }

    private void dataItemSelectedChanged(Data item) {
        item.textNode.setVisible(item.getSelected());
        updateDataItemStyle(item);
    }

    private Text createDataItemLabel(Data item) {
        Text text = item.textNode;
        text.setText(item.getName());
        return text;
    }

    private Node createNewDataItemSymbol(Data item) {
        Node shape = item.getNode();

        if(shape == null) {
            shape = new StackPane();
            item.setNode(shape);
        }

        updateDataItemStyle(item);

        return shape;
    }

    private void unitVectorAdded(UnitVector vector) {
        vector.setChart(StarCoordinatesChart.this);

        if(vector.getXValue() == 0.0 && vector.getYValue() == 0.0) {
            // Generate new position if vector contains default values
            generateNewPosition(vector);
        }

        updateAllDataValues();

        Node shape = createNewUnitVectorSymbol(vector);
        Text text = createUnitVectorLabel(vector);
        makeDraggable(vector);

        vector.vectorLine.getStyleClass().addAll("chart-unit-vector", "chart-unit-vector-line");

        getChartChildren().add(vector.vectorLine);
        getChartChildren().add(shape);
        getChartChildren().add(text);
        orderChartChildren();   // Make sure text and unit vectors are on top

        if(shouldAnimate()) {
            // TODO
        }
    }

    /**
     * Repositions the vector to a random position with
     * {@link StarCoordinatesChart#DEFAULT_UNIT_VECTOR_LENGTH}
     * distance to the center.
     */
    private void generateNewPosition(UnitVector vector) {
        Point2D position = new Point2D(ThreadLocalRandom.current().nextDouble(),
                                       ThreadLocalRandom.current().nextDouble())
                .normalize().multiply(DEFAULT_UNIT_VECTOR_LENGTH);

        vector.setXValue(position.getX());
        vector.setYValue(position.getY());
    }

    private Node createNewUnitVectorSymbol(UnitVector vector) {
        Node shape = vector.getNode();

        if(shape == null) {
            shape = new StackPane();
            vector.setNode(shape);
        }

        shape.getStyleClass().setAll("chart-unit-vector", "chart-symbol");

        return shape;
    }

    private Text createUnitVectorLabel(UnitVector vector) {
        Text text = vector.textNode;
        text.setText(vector.getName());
        return text;
    }

    private void makeDraggable(UnitVector vector) {
        final Delta dragDelta = new Delta();
        Node node = vector.getNode();
        Node vectorLine = vector.vectorLine;

        node.setOnMouseEntered(me -> node.getScene().setCursor(Cursor.HAND));
        node.setOnMouseExited(me -> node.getScene().setCursor(Cursor.DEFAULT));
        vectorLine.setOnMouseEntered(me -> node.getScene().setCursor(Cursor.HAND));
        vectorLine.setOnMouseExited(me -> node.getScene().setCursor(Cursor.DEFAULT));

        node.setOnMousePressed(me -> {
            if(me.isPrimaryButtonDown()) {
                node.getScene().setCursor(Cursor.CLOSED_HAND);

                dragDelta.x = deScaleValue(me.getSceneX()) - vector.getXValue();
                dragDelta.y = deScaleValue(me.getSceneY()) - vector.getYValue();
            }
        });

        node.setOnMouseDragged(me -> {
            if(me.isPrimaryButtonDown()) {
                vector.setXValue(deScaleValue(me.getSceneX()) - dragDelta.x);
                vector.setYValue(deScaleValue(me.getSceneY()) - dragDelta.y);
            }
        });

        vectorLine.setOnMousePressed(me -> {
            if(me.isPrimaryButtonDown()) {
                vectorLine.getScene().setCursor(Cursor.CLOSED_HAND);

                dragDelta.x = deScaleValue(me.getSceneX()) - vector.getXValue();
                dragDelta.y = deScaleValue(me.getSceneY()) - vector.getYValue();
            }
        });

        vectorLine.setOnMouseDragged(me -> {
            if(me.isPrimaryButtonDown()) {
                vectorLine.getScene().setCursor(Cursor.CLOSED_HAND);

                double deltaX = deScaleValue(me.getSceneX()) - dragDelta.x;
                double deltaY = deScaleValue(me.getSceneY()) - dragDelta.y;
                Point2D mouseDirection = new Point2D(deltaX, deltaY);
                Point2D vectorDirection = new Point2D(vector.getXValue(), vector.getYValue());
                Point2D projection = project(mouseDirection, vectorDirection);

                vector.setXValue(projection.getX());
                vector.setYValue(projection.getY());
            }
        });
    }

    /**
     * Projects vector a onto vector b.
     *
     * Creates a vector with a's "magnitude in the direction of vector b".
     *
     * Example:
     * a = (4, 3), b = (0, 10),
     * projection = (0, 3)
     */
    private static Point2D project(Point2D a, Point2D b) {
        return b.multiply(
                a.dotProduct(b) / Math.pow(b.magnitude(), 2));
    }

    /**
     * Sort the nodes so that UnitVector arrows are on top,
     * then Text and UnitVector lines and lastly other nodes.
     */
    private void orderChartChildren() {
        ObservableList<Node> chartChildren = getChartChildren();

        // Vector lines
        for(int i = 0; i < chartChildren.size(); i++) {
            Node node = chartChildren.get(i);
            if(node instanceof Path) {
                node.toFront();
            }
        }

        // UnitVector symbol
        for(int i = 0; i < chartChildren.size(); i++) {
            Node node = chartChildren.get(i);
            if(node.getStyleClass().contains("chart-unit-vector")) {
                node.toFront();
            }
        }

        // Text
        for(int i = 0; i < chartChildren.size(); i++) {
            Node node = chartChildren.get(i);
            if(node instanceof Text) {
                node.toFront();
            }
        }
    }

    private void dataStyleChanged(Data item) {
        updateDataItemStyle(item);
        requestChartLayout();
    }

    private void updateDataItemStyle(Data item) {
        Node node = item.getNode();
        if(node != null) {
            node.getStyleClass().setAll("chart-data-point", "chart-symbol");

            // Selected mode on or off
            if(item.getSelected()) {
                node.getStyleClass().add("chart-symbol-selected");
            }

            if(item.getStyleClasses() != null) {
                node.getStyleClass().addAll(item.getStyleClasses());
            }
        }
    }

    private void unitVectorNameChanged(UnitVector vector) {
        vector.textNode.setText(vector.getName());
        requestChartLayout();
    }

    public final static class UnitVector {

        private Text textNode = new Text();
        private Path vectorLine = new Path();

        /**
         * Readonly access to the node that represents the pie slice. You can use this to add mouse event listeners etc.
         */
        private ReadOnlyObjectWrapper<Node> node = new ReadOnlyObjectWrapper<>(this, "node");

        /**
         * The chart which this data belongs to.
         */
        private ReadOnlyObjectWrapper<StarCoordinatesChart> chart = new ReadOnlyObjectWrapper<>(this, "chart");

        /**
         * The name of the unit vector
         */
        private StringProperty name = new StringPropertyBase() {
            @Override
            protected void invalidated() {
                if(getChart() != null) getChart().unitVectorNameChanged(UnitVector.this);
            }

            @Override
            public Object getBean() {
                return UnitVector.this;
            }

            @Override
            public String getName() {
                return "name";
            }


        };
        private DoubleProperty xValue = new DoublePropertyBase() {

            @Override
            protected void invalidated() {
                if(getChart() != null) {
                    getChart().unitVectorChanged(UnitVector.this);
                }
            }

            @Override
            public Object getBean() {
                return UnitVector.this;
            }

            @Override
            public String getName() {
                return "xValue";
            }
        };
        private DoubleProperty yValue = new DoublePropertyBase() {

            @Override
            protected void invalidated() {
                if(getChart() != null) {
                    getChart().unitVectorChanged(UnitVector.this);
                }
            }

            @Override
            public Object getBean() {
                return UnitVector.this;
            }

            @Override
            public String getName() {
                return "yValue";
            }
        };
        private volatile boolean isBeingMovedByHand = false;

        public UnitVector(String name) {
            setName(name);
        }

        public final java.lang.String getName() {
            return name.getValue();
        }

        public final void setName(java.lang.String value) {
            name.setValue(value);
        }

        public final StringProperty nameProperty() {
            return name;
        }

        /**
         * Returns the node that represents the unit vector on the chart. You can use this to
         * add mouse event listeners etc.
         */
        public Node getNode() {
            return node.getValue();
        }

        private void setNode(Node value) {
            node.setValue(value);
        }

        public ReadOnlyObjectProperty<Node> nodeProperty() {
            return node.getReadOnlyProperty();
        }

        public double getXValue() {
            return xValue.getValue();
        }

        public void setXValue(double value) {
            xValue.setValue(value);
        }

        public DoubleProperty xValueProperty() {
            return xValue;
        }

        public double getYValue() {
            return yValue.getValue();
        }

        public void setYValue(double value) {
            yValue.setValue(value);
        }

        public DoubleProperty yValueProperty() {
            return yValue;
        }

        public StarCoordinatesChart getChart() {
            return chart.getValue();
        }

        private void setChart(StarCoordinatesChart value) {
            chart.setValue(value);
        }

        public final ReadOnlyObjectProperty<StarCoordinatesChart> chartProperty() {
            return chart.getReadOnlyProperty();
        }
    }

    public final static class Data {

        private Text textNode = new Text();

        /**
         * The chart which this data belongs to.
         */
        private ReadOnlyObjectWrapper<StarCoordinatesChart> chart = new ReadOnlyObjectWrapper<>(this, "chart");

        /**
         * Readonly access to the node that represents the pie slice. You can use this to add mouse event listeners etc.
         */
        private ReadOnlyObjectWrapper<Node> node = new ReadOnlyObjectWrapper<>(this, "node");

        /**
         * List of style classes that the
         */
        private ListProperty<String> styleClasses = new ListPropertyBase<String>(FXCollections.observableArrayList()) {

            @Override
            protected void invalidated() {
                if(getChart() != null) {
                    getChart().dataStyleChanged(Data.this);
                }
            }

            @Override
            public Object getBean() {
                return Data.this;
            }

            @Override
            public String getName() {
                return "styleClasses";
            }
        };

        /**
         * The name of the Data item
         */
        private StringProperty name = new StringPropertyBase() {
            @Override
            protected void invalidated() {
                if(getChart() != null) getChart().dataNameChanged(Data.this);
            }

            @Override
            public Object getBean() {
                return Data.this;
            }

            @Override
            public String getName() {
                return "name";
            }
        };

        private DoubleProperty xValue = new DoublePropertyBase() {

            @Override
            public Object getBean() {
                return Data.this;
            }

            @Override
            public String getName() {
                return "xValue";
            }
        };

        private DoubleProperty yValue = new DoublePropertyBase() {

            @Override
            public Object getBean() {
                return Data.this;
            }

            @Override
            public String getName() {
                return "yValue";
            }
        };

        /**
         * Map of attributes and their values for this object.
         */
        private MapProperty<String, Double> attributeValues = new MapPropertyBase<String, Double>() {

            @Override
            protected void invalidated() {
                if(getChart() != null) {
                    getChart().dataValueChanged(Data.this);
                }
            }

            @Override
            public Object getBean() {
                return Data.this;
            }

            @Override
            public String getName() {
                return "attributeValues";
            }
        };

        /**
         * Indicates that the Data item has been selected.
         */
        private BooleanProperty selected = new BooleanPropertyBase() {

            @Override
            protected void invalidated() {
                if(getChart() != null) {
                    getChart().dataItemSelectedChanged(Data.this);
                }
            }

            @Override
            public Object getBean() {
                return Data.this;
            }

            @Override
            public String getName() {
                return "selected";
            }
        };

        public Data(String name, ObservableMap<String, Double> attributeValues,
                    String... styleClasses) {
            textNode.setVisible(false);
            setName(name);
            setAttributeValues(attributeValues);
            getStyleClasses().addAll(styleClasses);
        }

        public final boolean getSelected() {
            return selected.get();
        }

        public final void setSelected(boolean value) {
            selected.setValue(value);
        }

        public final BooleanProperty selectedProperty() {
            return selected;
        }

        public ObservableList<String> getStyleClasses() {
            return styleClasses.getValue();
        }

        public void setStyleClasses(ObservableList<String> value) {
            styleClasses.setValue(value);
        }

        public final java.lang.String getName() {
            return name.getValue();
        }

        public final void setName(java.lang.String value) {
            name.setValue(value);
        }

        public final StringProperty nameProperty() {
            return name;
        }

        public ObservableMap<String, Double> getAttributeValues() {
            return attributeValues.getValue();
        }

        public void setAttributeValues(ObservableMap<String, Double> value) {
            attributeValues.setValue(value);
        }

        public MapProperty<String, Double> attributeValuesProperty() {
            return attributeValues;
        }

        public ListProperty<String> styleClassesProperty() {
            return styleClasses;
        }

        public double getXValue() {
            return xValue.getValue();
        }

        public void setXValue(double value) {
            xValue.setValue(value);
        }

        public DoubleProperty xValueProperty() {
            return xValue;
        }

        public double getYValue() {
            return yValue.getValue();
        }

        public void setYValue(double value) {
            yValue.setValue(value);
        }

        public DoubleProperty yValueProperty() {
            return yValue;
        }

        public StarCoordinatesChart getChart() {
            return chart.getValue();
        }

        private void setChart(StarCoordinatesChart value) {
            chart.setValue(value);
        }

        public final ReadOnlyObjectProperty<StarCoordinatesChart> chartProperty() {
            return chart.getReadOnlyProperty();
        }

        /**
         * Returns the node that represents the data point on the chart. You can use this to
         * add mouse event listeners etc.
         */
        public Node getNode() {
            return node.getValue();
        }

        private void setNode(Node value) {
            node.setValue(value);
        }

        public ReadOnlyObjectProperty<Node> nodeProperty() {
            return node.getReadOnlyProperty();
        }
    }

    private static class Delta {
        public double x;
        public double y;
    }
}
